"""
.. module:: testsequencer
    :platform: Darwin, Linux, Unix, Windows
    :synopsis: Module containing the base :class:`TestSequencer` type which is use to control
        the flow of the automation environment startup and test execution sequence.

.. moduleauthor:: Myron Walker <myron.walker@gmail.com>
"""

__author__ = "Myron Walker"
__copyright__ = "Copyright 2020, Myron W Walker"
__credits__ = []
__version__ = "1.0.0"
__maintainer__ = "Myron Walker"
__email__ = "myron.walker@gmail.com"
__status__ = "Development" # Prototype, Development or Production
__license__ = "MIT"

from typing import Any, List, OrderedDict, Sequence, Tuple

import collections
import json
import os
import sys
import uuid

from akit.environment.context import ContextUser
from akit.exceptions import AKitRuntimeError, AKitSemanticError, AKitSkipError
from akit.exceptions import format_exception
from akit.xlogging.foundations import getAutomatonKitLogger
from akit.xconfiguration import DiagnosticLabel, RuntimeConfigPaths

from akit.jsos import CHAR_RECORD_SEPERATOR
from akit.paths import get_path_for_diagnostics, get_path_for_output
from akit.results import ResultCode, ResultContainer, ResultNode, ResultType
from akit.compat import import_file
from akit.recorders import ResultRecorder

from akit.testing.constraints import Constraints
from akit.testing.testplus.testcollector import TestCollector
from akit.testing.testplus.registration.resourceregistry import resource_registry
from akit.testing.testplus.testgroup import TestGroup
from akit.testing.testplus.testref import TestRef
from akit.testing.testplus.markers import MetaFilter

from akit.timemachine import WELLKNOWN_PORTALS, timemachine_timeportal_code_append
from akit.xdebugger import WELLKNOWN_BREAKPOINTS, debugger_wellknown_breakpoint_code_append

logger = getAutomatonKitLogger()

CONSTRAINT_IMPORT_INSERTION_POINT = "# ------- INSERT CONSTRAINT IMPORTS HERE -------"

FACTORY_IMPORT_INSERTION_POINT = "# ------- INSERT FACTORY IMPORTS HERE -------"

TEMPLATE_TESTRUN_SEQUENCE_MODULE = '''"""
    ======================================= CODE GENERATED - DO NOT EDIT =======================================
    This is a code generated execution sequence document, do not manually edit this document.  The 'testplus'
    test framework generates this document in order to layout the test run scopes, parameter creations, and
    test calls in a user ledgible way which is easy to read and also to debug.
"""

__traceback_format_policy__ = "Brief"

from akit.xlogging.foundations import getAutomatonKitLogger

logger = getAutomatonKitLogger()

{}

{}

'''.format(CONSTRAINT_IMPORT_INSERTION_POINT, FACTORY_IMPORT_INSERTION_POINT)



class TEST_SEQUENCER_PHASES:
    """
        Indicates the current state of the sequencer.
    """
    Initial = 0
    Discovery = 1
    Collection = 2
    Graph = 3
    Traversal = 4

class SequencerScopeBase:

    def __init__(self, sequencer: "TestSequencer", recorder: ResultRecorder):
        super().__init__()

        self._sequencer = sequencer
        self._recorder = recorder
        return

    def _mark_descendants_as_error(self, cursor, cursor_id, errmsg):

        for chkey in cursor.children:
            child = cursor.children[chkey]
            if not child.finalized:
                if isinstance(child, TestRef):
                    self._mark_test_as_error(child, cursor_id, errmsg)
                elif isinstance(child, TestGroup):
                    scope_id = str(uuid.uuid4())
                    scope_name = child.scope_name
                    result = self._sequencer.create_test_result_container(scope_id, scope_name, parent_inst=cursor_id)
                    self._recorder.record(result)
                    self._mark_descendants_as_error(child, scope_id, errmsg)

        return

    def _mark_descendants_skipped(self, cursor, cursor_id, reason, bug):

        for chkey in cursor.children:
            child = cursor.children[chkey]
            if not child.finalized:
                if isinstance(child, TestRef):
                    self._mark_test_as_skip(child, cursor_id, reason, bug)
                elif isinstance(child, TestGroup):
                    scope_id = str(uuid.uuid4())
                    scope_name = child.scope_name
                    result = self._sequencer.create_test_result_container(scope_id, scope_name, parent_inst=cursor_id)
                    self._recorder.record(result)
                    self._mark_descendants_skipped(child, scope_id, reason, bug)

        return

    def _mark_test_as_error(self, testref: TestRef, parent_id: str, errmsg: str):
        test_id = str(uuid.uuid4())
        result = self._sequencer.create_test_result_node(test_id, testref.test_name, testref.monikers, testref.pivots, parent_inst=parent_id)
        result.add_error(errmsg)
        result.finalize()
        self._recorder.record(result)
        return

    def _mark_test_as_skip(self, testref: TestRef, parent_id: str, reason: str, bug: str):
        test_id = str(uuid.uuid4())
        result = self._sequencer.create_test_result_node(test_id, testref.test_name, testref.monikers, testref.pivots, parent_inst=parent_id)
        result.mark_skip(reason, bug)
        result.finalize()
        self._recorder.record(result)
        return

class SequencerModuleScope(SequencerScopeBase):
    def __init__(self, sequencer: "TestSequencer", recorder: ResultRecorder, scope_name: str, **kwargs):
        super().__init__(sequencer, recorder)

        self._scope_name = scope_name
        self._scope_args = kwargs
        self._scope_id = None
        self._parent_scope_id = None
        self._scope_node = self._sequencer.find_treenode_for_scope(scope_name)
        return

    def __enter__(self):
        self._parent_scope_id, self._scope_id = self._sequencer.scope_id_create(self._scope_name)
        result = self._sequencer.create_test_result_container(self._scope_id, self._scope_name, parent_inst=self._parent_scope_id)
        self._recorder.record(result)
        logger.info("MODULE ENTER: {}, {}".format(self._scope_name, self._scope_id))
        return self

    def __exit__(self, ex_type, ex_inst, ex_tb):
        handled = False

        if ex_type is not None:

            if issubclass(ex_type, AKitSkipError):
                self._mark_descendants_skipped(self._scope_node, self._scope_id, ex_inst.reason, ex_inst.bug)
            else:
                errmsg = ""
                self._mark_descendants_as_error(self._scope_node,  self._scope_id, errmsg)

            # If an exceptions was thrown in this context, it means
            # that the exception occured during the setup for this
            # module, this means we need to mark all descendant tests
            # as error'd due to a setup failure.
            errmsg = "Exception raises setting up scope='{}'".format(self._scope_name)
            logger.exception(errmsg)
            handled = True

        self._scope_node.finalize()
        self._sequencer.scope_id_pop(self._scope_name)
        logger.info("MODULE EXIT: {}, {}".format(self._scope_name, self._scope_id))
        return handled

    

class SequencerSessionScope(SequencerScopeBase):
    def __init__(self, sequencer: "TestSequencer", recorder: ResultRecorder, root_result:ResultContainer):
        super().__init__(sequencer, recorder)
        
        self._scope_name = root_result.name
        self._scope_id = root_result.inst_id
        self._root_result = root_result
        self._scope_node = self._sequencer.testtree
        return

    def __enter__(self):
        self._sequencer.scope_id_push(self._scope_name, self._scope_id)
        logger.info("SESSION ENTER: {}".format(self._scope_id))
        return self

    def __exit__(self, ex_type, ex_inst, ex_tb):
        handled = False

        if ex_type is not None:

            if issubclass(ex_type, AKitSkipError):
                self._mark_descendants_skipped(self._scope_node, self._scope_id, ex_inst.reason, ex_inst.bug)
            else:
                errmsg = ""
                self._mark_descendants_as_error(self._scope_node,  self._scope_id, errmsg)

            # If an exceptions was thrown in this context, it means
            # that the exception occured during the setup for this
            # module, this means we need to mark all descendant tests
            # as error'd due to a setup failure.
            errmsg = "Exception raises setting up scope='{}'".format(self._scope_name)
            logger.exception(errmsg)
            handled = True

        self._scope_node.finalize()
        self._sequencer.scope_id_pop(self._scope_name)
        logger.info("SESSION EXIT: {}".format(self._scope_id))
        return handled

class SequencerTestScope:
    def __init__(self, sequencer: "TestSequencer", recorder: ResultRecorder, test_name: str, parameterized: OrderedDict[str, Any]={}):
        super().__init__()

        self._sequencer = sequencer
        self._recorder = recorder
        self._test_name = test_name
        self._scope_id = None
        self._parent_scope_id = None
        self._result = None
        self._scope_node = self._sequencer.find_treenode_for_scope(test_name)
        self._parameterized = parameterized
        self._monikers, self._pivots = self._get_monikers_and_pivots()
        self._context_identifier = "{}:{}".format(self._test_name, ",".join(self._monikers))
        return

    def _get_monikers_and_pivots(self) -> Tuple[List[str], OrderedDict[str, Any]]:
        """
            Creates a full context identifier based on the full testname and the identifiers of any
            parameterized parameters that are being passed to the test.
        """
        
        monikers = []
        pivots = collections.OrderedDict()

        if len(self._parameterized) > 0:

            for pname, pobj in self._parameterized.items():
                
                if hasattr(pobj, "moniker"):
                    monikers.append(pobj.moniker)
                else:
                    monikers.append(str(pobj))
                
                if hasattr(pobj, "pivots"):
                    pivots[pname] = pobj.pivots
                else:
                    pivots[pname] = str(pobj)

        return monikers, pivots

    def __enter__(self):
        self._parent_scope_id, self._scope_id = self._sequencer.scope_id_create(self._context_identifier)
        logger.info("TEST SCOPE ENTER: {}, {}".format(self._context_identifier, self._scope_id))
        self._result = self._sequencer.create_test_result_node(self._scope_id, self._test_name, self._monikers, self._pivots, parent_inst=self._parent_scope_id)
        self._test_scope_enter()
        return self

    def __exit__(self, ex_type, ex_inst, ex_tb):
        handled = True

        if ex_type is not None:

            if issubclass(ex_type, AKitSkipError):
                self._result.mark_skip(ex_inst.reason, ex_inst.bug)
            else:
                # If an exceptions was thrown in this context, it means
                # that a test threw an exception.
                ex_lines = format_exception(ex_inst)
                
                if issubclass(ex_type, AssertionError):
                    # The convention for test failures that all tests should throw
                    # an AssertionError derived exception for failure conditions.
                    # This is important because a failure condition implies an expectation
                    # was checked and not met which implies a product code related failure
                    
                    self._result.add_failure(ex_lines)
                else:
                    self._result.add_error(ex_lines)

                errmsg_lines = [
                    "Exception raises setting up scope='{}'".format(self._test_name)
                ]
                for exln in ex_lines:
                    errmsg_lines.append(exln.rstrip())

                errmsg = os.linesep.join(errmsg_lines)
                logger.error(errmsg)

            handled = True
        else:
            self._result.mark_passed()

        # Call test scope exit before we finalize our results
        self._test_scope_exit()

        self._result.finalize()
        self._recorder.record(self._result)

        self._scope_node.finalize()
        self._sequencer.scope_id_pop(self._context_identifier)

        logger.info("TEST SCOPE EXIT: {}, {}".format(self._context_identifier, self._scope_id))

        return handled

    def _test_scope_enter(self):
        return
    
    def _test_scope_exit(self):
        return

class SequencerTestSetupScope:
    def __init__(self, sequencer, recorder, test_name, **kwargs):
        super().__init__()

        self._sequencer = sequencer
        self._recorder = recorder
        self._test_name = test_name
        self._scope_name = "setup:{}".format(test_name)
        self._scope_args = kwargs
        self._scope_id = None
        self._parent_scope_id = None
        self._test_scope_node = self._sequencer.find_treenode_for_scope(test_name)
        return

    def __enter__(self):
        self._parent_scope_id, self._scope_id = self._sequencer.scope_id_create(self._scope_name)
        logger.info("TEST SETUP ENTER: {}, {}".format(self._scope_name, self._scope_id))
        return self

    def __exit__(self, ex_type, ex_inst, ex_tb):
        handled = False

        if ex_type is not None:
            # If an exceptions was thrown in this context, it means
            # that the exception occured during the setup for this
            # module, this means we need to mark all descendant tests
            # as error'd due to a setup failure.
            errmsg = "Exception raises setting up scope='{}'".format(self._scope_name)
            logger.exception(errmsg)
            handled = True

        self._test_scope_node.finalize()
        self._sequencer.scope_id_pop(self._scope_name)
        logger.info("TEST SETUP EXIT: {}, {}".format(self._scope_name, self._scope_id))
        return handled

class TestSequencer(ContextUser):
    """
        The :class:`TestSequencer` is a state machine that helps to orchestrate the flow fo the test run.  It ensures
        that the steps of the test flow are consistent between runs.
    """

    def __init__(self, jobtitle: str, root: str, includes: Sequence[str], excludes: Sequence[str], metafilters: List[MetaFilter]):
        """
            Creates a 'TestSequencer' object which is used to discover the tests and control the flow of a test run.

            :param jobtitle: The name of the test job.
            :param root: The path to the root folder that is the base of the tests.
            :param includes: List of expressions used to determine which tests to include.
                             (scope):(package).(package)@(module)#(testname)
            :param excludes: List of expressions used to determine which tests to exclued from the included tests.
            :param metafilters: List of expressions used to filter tests by metadata
        """
        super().__init__()

        self._jobtitle = jobtitle
        self._root = root
        self._includes = includes
        self._excludes = excludes
        self._metafilters = metafilters
        self._integrations = {}
        self._references = []
        self._scopes = {}
        self._scope_roots = []
        self._import_errors = {}
        self._testtree = None
        self._landscape = None
        self._sequence_document = None
        self._recorder = None
        self._root_result = None
        self._scope_stack = []
        return

    def __enter__(self):
        """
            Provides 'with' statement scope semantics for the :class:`TestSequencer`
        """
        return self

    def __exit__(self, ex_type, ex_inst, ex_tb):
        """
            Provides 'with' statement scope semantics for the :class:`TestSequencer`
        """
        return False

    @property
    def import_errors(self):
        """
            A list of import errors that were encountered during the sequencing of the test run.
        """
        return self._import_errors

    @property
    def references(self):
        """
            A list of :class:`TestRef` objects that are included in the test run.
        """
        return self._references

    @property
    def testnodes(self):
        """
        """
        root_nodes = [tnode for _, tnode in self._testtree]
        return root_nodes

    @property
    def testtree(self):
        """
        """
        return self._testtree

    def attach_to_framework(self, landscape):
        """
            Goes through all the integrations and provides them with an opportunity to
            attach to the test environment.
        """

        self._landscape = landscape

        for _, integ_type in self._integrations.items():
            integ_type.attach_to_framework(landscape)

        return

    def attach_to_environment(self, landscape):
        """
            Goes through all the integrations and provides them with an opportunity to
            attach to the test environment.
        """

        results_dir = get_path_for_output()

        environment_dict = self.create_password_masked_environment()

        package_dict = collections.OrderedDict()

        package_names = [k for k in sys.modules.keys()]
        package_names.sort()
        for pname in package_names:
            nxtmod = sys.modules[pname]
            if pname.find(".") == -1 and hasattr(nxtmod, "__file__"):
                package_dict[pname] = nxtmod.__file__

        startup_dict = collections.OrderedDict([
            ("command", " ".join(sys.argv)),
            ("environment", environment_dict),
            ("packages", package_dict)
        ])

        startup_full = os.path.join(results_dir, "startup-configuration.json")
        with open(startup_full, 'w') as suf:
            json.dump(startup_dict, suf, indent=True)

        for _, integ_type in self._integrations.items():
            integ_type.attach_to_environment()

        return

    def collect_resources(self):
        """
            Goes through all the integrations and provides them with an opportunity to
            collect shared resources that are required for testing.
        """

        for _, integ_type in self._integrations.items():
            integ_type.collect_resources()

        return

    def create_job_result_container(self, scope_id: str, scope_name: str) -> ResultContainer:
        """
            Method for creating a result container.
        """
        rcontainer = ResultContainer(scope_id, scope_name, ResultType.JOB, parent_inst=None)
        return rcontainer

    def create_password_masked_environment(self):
        """
            Creates copy of the environment for writing to a file with the passwords masked.
        """
        environment_dict = collections.OrderedDict()

        env_keys = [k for k in os.environ.keys()]
        env_keys.sort()
        for ek in env_keys:
            environment_dict[ek] = os.environ[ek]

        for key in environment_dict.keys():
            if key.find("PASSWORD") > -1:
                environment_dict[key] = "(hidden)"
        
        return environment_dict

    def create_test_result_container(self, scope_id: str, scope_name: str, parent_inst: str) -> ResultContainer:
        """
            Method for creating a result container.
        """
        rcontainer = ResultContainer(scope_id, scope_name, ResultType.TEST_CONTAINER, parent_inst=parent_inst)
        return rcontainer
    
    def create_test_result_node(self, scope_id: str, name: str, monikers: List[str], pivots: OrderedDict[str, Any], parent_inst: str) -> ResultNode:
        """
            Method for creating a result node.
        """
        rnode = ResultNode(scope_id, name, monikers, pivots, ResultType.TEST, parent_inst=parent_inst)
        return rnode

    def diagnostic_capture_pre_testrun(self, level: int=9):
        """
            Perform a pre-run diagnostic on the devices in the test landscape.
        """

        prerun_info = self.context.lookup(RuntimeConfigPaths.DIAGNOSTIC_PRERUN)

        if prerun_info is not None:
            label = DiagnosticLabel.PRERUN_DIAGNOSTIC
            diagnostic_root = get_path_for_diagnostics(label)

            for _, integ_type in self._integrations.items():
                integ_type.diagnostic(label, level, diagnostic_root)

        return

    def diagnostic_capture_post_testrun(self, level: int=9):
        """
            Perform a post-run diagnostic on the devices in the test landscape.
        """
        postrun_info = self.context.lookup(RuntimeConfigPaths.DIAGNOSTIC_POSTRUN)

        if postrun_info is not None:

            label = DiagnosticLabel.POSTRUN_DIAGNOSTIC
            diagnostic_root = get_path_for_diagnostics(label)

            for _, integ_type in self._integrations.items():
                integ_type.diagnostic(label, level, diagnostic_root)

        return

    def discover(self, test_module=None, include_integrations: bool=True):
        """
            Initiates the discovery phase of the test run.
        """
        collector = TestCollector(self._root, excludes=self._excludes, metafilters=self._metafilters, test_module=test_module)

        # Discover the tests, integration points, and scopes.  If test modules is not None then
        # we are running tests from an individual module and we can limit discovery to the test module
        for inc_item in self._includes:
            collector.collect_references(inc_item)

        collector.finalize_collection()

        self._testtree = collector.test_tree
        self._references = collector.references

        testcount = len(self._references)
        if testcount > 0:
            if include_integrations:
                self._integrations, self._scopes = collector.collect_integrations_and_scopes()
            self._import_errors = collector.import_errors

        return testcount

    def enter_module_scope_context(self, scope_name):
        context = SequencerModuleScope(self, self._recorder, scope_name)
        return context

    def enter_session_scope_context(self):
        context = SequencerSessionScope(self, self._recorder, self._root_result)
        return context

    def enter_test_scope_context(self, test_name, parameterized: Sequence=[]):
        context = SequencerTestScope(self, self._recorder, test_name, parameterized=parameterized)
        return context

    def enter_test_setup_context(self, test_name):
        context = SequencerTestSetupScope(self, self._recorder, test_name)
        return context

    def establish_integration_order(self):
        """
            Re-orders the integrations based on any declared precedences.
        """

        for _, integ_type in self._integrations.items():
            integ_type.establish_integration_order()

        return

    def establish_presence(self):
        """
            Goes through all the integrations and provides them with an opportunity to
            establish a presence or persistant services with the test resource or resources
            they are integrating into the automation run.
        """

        for _, integ_type in self._integrations.items():
            integ_type.establish_presence()

        return

    def execute_tests(self, runid: str, recorder):
        """
            Called in order to execute the tests contained in the :class:`TestPacks` being run.
        """

        res_name = "<session>"

        self._recorder = recorder

        self._root_result = self.create_job_result_container(runid, res_name)
        recorder.record(self._root_result)

        if self._sequence_document is None:
            errmsg = "The 'execute_tests' method should not be called without first generating the test sequence document."
            raise AKitSemanticError(errmsg) from None

        # Import the test sequence document
        sequence_mod = import_file("sequence_document", self._sequence_document)
        sequence_mod.session(self)

        return

    def expand_test_tree_based_on_query(self):

        context_locals = {}

        # Walk the test tree and expand the parameters
        # self._testtree

        return

    def find_treenode_for_scope(self, scope_name, cursor=None):
        found = None

        if cursor is None:
            cursor = self._testtree

        if cursor.scope_name == scope_name:
            found = cursor
        elif isinstance(cursor, TestGroup):
            for child in cursor.children.values():
                found = self.find_treenode_for_scope(scope_name, cursor=child)
                if found is not None:
                    break

        return found

    def generate_testrun_sequence_document(self, outputfilename: str, indent_space="    "):

        factory_imports = set()
        constraint_imports = set()

        temp_outputfile = outputfilename + ".tmp"

        with open(temp_outputfile, 'w') as sdf:
            current_node = self._testtree

            sdf.write(TEMPLATE_TESTRUN_SEQUENCE_MODULE)

            scopes_called = self._generate_session_method(sdf, self._testtree, factory_imports, constraint_imports, indent_space)

            while len(scopes_called) > 0:
                sdf.write(os.linesep)
                scope_module, scope_name, scope_node, scope_call_args = scopes_called.pop(0)
                child_scopes_called = self._generate_scope_method(sdf, scope_module, scope_name, scope_node, scope_call_args, factory_imports, constraint_imports, indent_space)
                child_scopes_called.extend(scopes_called)
                scopes_called = child_scopes_called

        linesep = os.linesep

        # Insert the factory imports and constraint imports
        with open(temp_outputfile, 'r') as tmpf:

            with open(outputfilename, 'w') as sdf:

                while True:
                    nxtline = tmpf.readline()
                    if len(nxtline) == 0:
                        break
                    
                    if nxtline.startswith(CONSTRAINT_IMPORT_INSERTION_POINT):
                        for cimp in constraint_imports:
                            sdf.write(cimp)
                            sdf.write(linesep)
                    elif nxtline.startswith(FACTORY_IMPORT_INSERTION_POINT):
                        for fimp in factory_imports:
                            sdf.write(fimp)
                            sdf.write(linesep)
                    else:
                        sdf.write(nxtline)

        os.remove(temp_outputfile)

        self._sequence_document = outputfilename

        return

    def scope_id_create(self, scope_name):

        parent_id = None
        if len(self._scope_stack) > 0:
            parent_id = self._scope_stack[-1][1]

        scope_id = str(uuid.uuid4())
        self._scope_stack.append((scope_name, scope_id))

        return parent_id, scope_id

    def scope_id_pop(self, scope_name):

        top_entry_name, _ = self._scope_stack.pop()
        if top_entry_name != scope_name:
            errmsg = "Attempting to pop '{}' from the scope stack but encountered '{}'.".format(scope_name, top_entry_name)
            raise AKitRuntimeError(errmsg) from None

        return

    def scope_id_push(self, scope_name, scope_id):
        self._scope_stack.append((scope_name, scope_id))
        return

    def _generate_session_method(self, outf, root_node, factory_imports, constraint_imports, indent_space):
        """
            Generates the :function:`session` entry point function or the test run
            sequence document.
        """
        scopes_called = []

        current_indent = indent_space

        method_lines = [
            'def session(sequencer):',
            '{}"""'.format(current_indent),
            '{}This is the entry point for the test run sequence document.'.format(indent_space),
            '{}"""'.format(current_indent)
        ]

        debugger_wellknown_breakpoint_code_append(WELLKNOWN_BREAKPOINTS.TESTRUN_START, method_lines, current_indent)

        current_indent += timemachine_timeportal_code_append(WELLKNOWN_PORTALS.TESTRUN_START, method_lines, current_indent)

        method_lines.append('')
        method_lines.append('{}with sequencer.enter_session_scope_context() as ssc:'.format(current_indent))
        method_lines.append('')

        current_indent = current_indent + indent_space

        this_scope = root_node.get_resource_scope()

        # Import all the parameter source functions
        for _, porigin in this_scope.parameter_originations.items():
            method_lines.append('{}from {} import {}'.format(current_indent, porigin.source_module_name, porigin.source_function.__name__))
        method_lines.append('')

        # Create the calls to all of the root test scopes
        child_call_args = ['sequencer']

        # Create the variables with session scope
        for pname, porigin in this_scope.parameter_originations.items():
            source_func_call = porigin.generate_call()
            method_lines.append('{}for {} in {}:'.format(current_indent, pname, source_func_call))
            child_call_args.append(pname)
            current_indent = current_indent + indent_space

        child_name_list = [cnk for cnk in root_node.children.keys()]
        child_name_list.sort()
        for child_name in child_name_list:
            child_node = root_node.children[child_name]
            scope_module = child_name
            if child_node.package:
                scope_module = child_node.package + "." + child_name
            scope_name = scope_module.replace(".", "_")
            call_line = '{}scope_{}({})'.format(current_indent, scope_name, ", ".join(child_call_args))
            method_lines.append(call_line)
            scopes_called.append((scope_module, scope_name, child_node, child_call_args))

        method_lines.append('')
        method_lines.append('{}return'.format(indent_space))
        method_lines.append('')

        method_content = os.linesep.join(method_lines)
        outf.write(method_content)

        return scopes_called

    def _generate_scope_method(self, outf, scope_module, scope_name, scope_node, scope_call_args, factory_imports, constraint_imports, indent_space):
        scopes_called = []

        current_indent = "    "

        method_lines = [
            'def scope_{}({}):'.format(scope_name, ", ".join(scope_call_args)),
            '{}"""'.format(current_indent),
            "{}This is the entry point for the '{}' test scope.".format(current_indent, scope_name),
            '{}"""'.format(current_indent),
            '',
            '{}with sequencer.enter_module_scope_context("{}") as msc:'.format(current_indent, scope_module)
        ]

        current_indent = current_indent + indent_space

        # Create the calls to all of the root test scopes
        child_call_args = [arg for arg in scope_call_args]

        resource_scope = scope_node.get_resource_scope()

        # Add the local factory imports to our imports list
        for _, porigin in resource_scope.parameter_originations.items():
            factory_imports.add('from {} import {}'.format(porigin.source_module_name, porigin.source_function.__name__))

        # Create the variables with session scope
        for pname, porigin in resource_scope.parameter_originations.items():

            sp_parameter_names = [pn for pn in porigin.source_signature.parameters.keys()]
            if 'constraints' in sp_parameter_names:
                constraints = {}
                if porigin.constraints is not None:
                    constraints = porigin.constraints
                    if issubclass(type(constraints), Constraints):
                        constraint_imports.add(constraints.get_import_statement())
                method_lines.append('{}constraints={}'.format(current_indent, repr(constraints)))
            
            source_func_call = porigin.generate_call()
            method_lines.append('{}for {} in {}:'.format(current_indent, pname, source_func_call))
            child_call_args.append(pname)
            current_indent = current_indent + indent_space

        child_name_list = [cnk for cnk in scope_node.children.keys()]
        child_name_list.sort()
        for child_name in child_name_list:
            child_node = scope_node.children[child_name]
            if isinstance(child_node, TestRef):
                pre_test_scope_indent = current_indent

                # Generate a test run scope for this test
                test_scope_name = child_node.test_name
                test_scope = resource_registry.lookup_resource_scope(test_scope_name)

                test_parameters = child_node.test_function_parameters

                test_local_args = []

                for param_name in test_parameters:
                    # If the test parameter is of test_scope origin
                    # then we need to add it to test_local_args so we
                    # can create an import for it.
                    if param_name in test_scope.parameter_originations:
                        param_obj = test_scope.parameter_originations[param_name]
                        test_local_args.append((param_name, param_obj))

                parameterized_args = ""

                method_lines.append("{}# ================ Test Scope: {} ================".format(current_indent, test_scope_name))
                method_lines.append('')
                # Import the test function and assign the test name
                method_lines.append("{}from {} import {}".format(current_indent, child_node.test_module_name, child_node.test_base_name))
                method_lines.append('{}test_scope_name = "{}"'.format(current_indent, test_scope_name))
                method_lines.append('')

                if len(test_local_args) > 0:
                    method_lines.append('{}with sequencer.enter_test_setup_context(test_scope_name) as tsetup:'.format(current_indent))
                    current_indent += indent_space

                    parameterized_args_names = []

                    # Import any factory functions that are used in test local factory methods
                    for param_name, param_obj in test_local_args:
                        fmodname = param_obj.source_module_name
                        ffuncname = param_obj.source_function_name
                        factory_imports.add("from {} import {}".format(fmodname, ffuncname))

                    method_lines.append('')

                    # Generate any local parameters
                    for param_name, param_obj in test_local_args:
                        ffuncname = param_obj.source_function_name
                        ffuncsig = param_obj.source_signature
                        ffuncargs = [pn for pn in ffuncsig.parameters]
                        if 'constraints' in ffuncargs:
                            constraints = {}
                            if param_obj.constraints is not None:
                                constraints = param_obj.constraints
                                if issubclass(type(constraints), Constraints):
                                    constraint_imports.add(constraints.get_import_statement())
                            method_lines.append('{}constraints={}'.format(current_indent, repr(constraints)))
                        ffuncargs_str = " ,".join(ffuncargs)
                        method_lines.append("{}for {} in {}({}):".format(current_indent, param_name, ffuncname, ffuncargs_str))
                        parameterized_args_names.append(param_name)
                        current_indent += indent_space
                
                    parameterized_args = ", ".join(map(lambda arg: "'{}': {}".format(arg, arg), parameterized_args_names))
                    if len(parameterized_args_names) == 1:
                        parameterized_args += ","

                    parameterized_args = ", parameterized={%s}" % (parameterized_args)

                    method_lines.append('')

                method_lines.append('{}with sequencer.enter_test_scope_context(test_scope_name{}) as tsc:'.format(current_indent, parameterized_args))
                current_indent += indent_space

                # Make the call to the test function
                test_args = []
                for param_name in test_parameters:
                    test_args.append(param_name)
                call_line = '{}{}({})'.format(current_indent, child_node.test_base_name, ", ".join(test_args))
                method_lines.append(call_line)
                method_lines.append('')

                # Restor the indent to before the test scope
                current_indent = pre_test_scope_indent

            else:
                scope_module = child_name
                if child_node.package:
                    scope_module = child_node.package + "." + child_name
                scope_name = scope_module.replace(".", "_")
                call_line = '{}scope_{}({})'.format(current_indent, scope_name, ", ".join(child_call_args))
                method_lines.append(call_line)
                scopes_called.append((scope_module, scope_name, child_node, child_call_args))

        method_lines.append('')
        method_lines.append('{}return'.format(indent_space))
        method_lines.append('')

        method_content = os.linesep.join(method_lines)
        outf.write(method_content)

        return scopes_called

    def record_import_errors(self, outputfilename: str):
        """
            Method that writes out the import errors to the active results directory.
        """
        with open(outputfilename, 'w') as ief:
            for modname, filename, errmsg in self._import_errors.values():
                ief.write(CHAR_RECORD_SEPERATOR)
                ieitem = {
                    "module": modname,
                    "filename": filename,
                    "trace": errmsg.splitlines(False)
                }
                json.dump(ieitem, ief, indent=4)

        return



"""
.. module:: landscape
    :platform: Darwin, Linux, Unix, Windows
    :synopsis: Module containing the Landscape related classes.

.. moduleauthor:: Myron Walker <myron.walker@gmail.com>

"""

__author__ = "Myron Walker"
__copyright__ = "Copyright 2020, Myron W Walker"
__credits__ = []
__version__ = "1.0.0"
__maintainer__ = "Myron Walker"
__email__ = "myron.walker@gmail.com"
__status__ = "Development" # Prototype, Development or Production
__license__ = "MIT"

from typing import Dict, List, Optional, Union

import inspect
import os
import pprint
import threading

import zeroconf
import zeroconf.const

from akit.compat import import_by_name

from akit.environment.context import Context
from akit.environment.variables import AKIT_VARIABLES

from akit.exceptions import AKitConfigurationError

from akit.interop.landscaping.layers.landscapeconfigurationlayer import LandscapeConfigurationLayer
from akit.interop.landscaping.layers.landscapeintegrationlayer import LandscapeIntegrationLayer
from akit.interop.landscaping.layers.landscapeoperationallayer import LandscapeOperationalLayer

from akit.interop.dns.mdnsserviceinfo import MdnsServiceInfo
from akit.interop.dns.mdnsservicecatalog import MdnsServiceCatalog

from akit.wellknown.singletons import LandscapeSingleton


from akit.xlogging.foundations import getAutomatonKitLogger

PASSWORD_MASK = "(hidden)"

def mask_passwords (context):
    """
        Takes a dictionary context object and will recursively mask any password members found
        in the dictionary.
    """
    for key, val in context.items():
        if (key == "password" or key == "secret"):
            context[key] = PASSWORD_MASK

        if isinstance(val, dict):
            mask_passwords(val)
        elif isinstance(val, list):
            for item in val:
                if isinstance(item, dict):
                    mask_passwords(item)

    return

def filter_credentials(device_info, credential_lookup, category):
    """
        Looks up the credentials associated with a device and returns the credentials found
        that match a given category.

        :param device_info: Device information dictionary with credential names to reference.
        :param credential_lookup: A credential lookup dictionary that is used to convert credential
                                  names into credential objects loaded from the landscape.
        :param category: The category of credentials to return when filtering credentials.
    """
    cred_found_list = []

    cred_name_list = device_info["credentials"]
    for cred_name in cred_name_list:
        if cred_name in credential_lookup:
            credential = credential_lookup[cred_name]
            if credential.category == category:
                cred_found_list.append(credential)
        else:
            error_lines = [
                "The credential '{}' was not found in the credentials list.",
                "DEVICE:"
            ]

            dev_repr_lines = pprint.pformat(device_info, indent=4).splitlines(False)
            for dline in dev_repr_lines:
                error_lines.append("    " + dline)

            error_lines.append("CREDENTIALS:")
            cred_available_list = [cname for cname in credential_lookup.keys()]
            cred_available_list.sort()
            for cred_avail in cred_available_list:
                error_lines.append("    " + cred_avail)

            errmsg = os.linesep.join(error_lines)
            raise AKitConfigurationError(errmsg) from None

    return cred_found_list

class Landscape(LandscapeConfigurationLayer, LandscapeIntegrationLayer, LandscapeOperationalLayer):
    """
        The base class for all derived :class:`Landscape` objects.  The :class:`Landscape`
        object is a singleton object that provides access to the resources and test
        environment level methods.  The functionality of the :class:`Landscape` object is setup
        so it can be transitioned through activation stages:
        
        * Configuration
        * Integration
        * Operational

        The different phases of operation of the landscape allow it to be used for a wider variety
        of purposes from commandline configuration and maintenance operations, peristent services
        and automation run functionality.

        The activation stages or levels of the :class:`Landscape` object are implemented using
        a python MixIn pattern in order to ensure that individual layers can be customized
        using object inheritance while at the same time keeping the object hierarchy simple.

        ..note: The :class:`Landscape` object constructor utilizes the `super` keyword for calling
        the mixin layer constructors using method resolution order or `mro`.  In order for `super`
        to work correctly all objects in the hierarchy should also provide a constructor and should
        also utilize `super`.  This is true also for objects that only inherit from :class:`object`.
        Should you need to create a custom layer override object, you must ensure the proper use
        of `super` in its constructor.
    """

    context = Context()

    logger = getAutomatonKitLogger()
    landscape_lock = threading.RLock()

    _landscape_type = None
    _instance = None
    _instance_initialized = False

    def __new__(cls):
        """
            Constructs new instances of the Landscape object from the :class:`Landscape`
            type or from a derived type that is found in the module specified in the
            :module:`akit.environment.variables` module or by setting the
            'AKIT_CONFIG_EXTENSION_POINTS_MODULE' environment variable and overloading
            the 'get_landscape_type' method.
        """

        if Landscape._landscape_type is None:
            if Landscape._instance is None:
                Landscape._instance = super(Landscape, cls).__new__(cls)
        elif Landscape._instance is None:
            Landscape._instance = super(Landscape, cls._landscape_type).__new__(cls._landscape_type)

        return Landscape._instance

    def __init__(self):
        """
            Creates an instance or reference to the :class:`Landscape` singleton object.  On the first call to this
            constructor the :class:`Landscape` object is initialized and the landscape configuration is loaded.
        """
        
        # We are a singleton so we only want the intialization code to run once
        Landscape.landscape_lock.acquire()
        if not Landscape._instance_initialized:
            Landscape._instance_initialized = True
            Landscape.landscape_lock.release()

            self._interactive_mode = False

            super().__init__()
        else:
            Landscape.landscape_lock.release()

        return

    @property
    def interactive_mode(self):
        """
            Returns a boolean indicating if interactive mode is on or off.
        """
        return self._interactive_mode

    @interactive_mode.setter
    def interactive_mode(self, interactive: bool) -> None:
        """
            Turn on or off interactive mode.
        """
        self._interactive_mode = interactive
        return

    @property
    def zeroconf(self) -> zeroconf.Zeroconf:
        """
            Returns the ZeroConf object.
        """
        return self._zeroconf
    
    @property
    def zeroconf_catalog(self) -> MdnsServiceCatalog:
        """
            Returns the service catalog that is used to lookup specific information about
            services reported by mDNS
        """
        return self._zeroconf_catalog

    @property
    def zeroconf_browser(self) -> "zeroconf.ServiceBrowser":
        """
            Returns the service browser used for searching for mDNS services.
        """
        return self._zeroconf_browser

    def mdns_list_service_names_for_type(self, svc_type: str) -> List[str]:

        svc_name_list = self._zeroconf_catalog.list_service_names_for_type(svc_type)

        return svc_name_list

    def mdns_lookup_service_info(self, svc_type: str, svc_name: str) -> MdnsServiceInfo:
        
        service_info = self._zeroconf_catalog.lookup_service_info(svc_type, svc_name)

        return service_info

def is_subclass_of_landscape(cand_type):
    """
        Returns a boolean value indicating if the candidate type is a subclass
        of :class:`Landscape`.
    """
    is_scol = False
    if inspect.isclass(cand_type) and issubclass(cand_type, Landscape):
        is_scol = True
    return is_scol

def load_and_set_landscape_type(lscape_module):
    """
        Scans the module provided for :class:`Landscape` derived classes and will
        take the first one and assign it as the current runtime landscape type.
    """
    class_items = inspect.getmembers(lscape_module, is_subclass_of_landscape)
    for _, cls_type in class_items:
        type_module_name = cls_type.__module__
        if type_module_name == lscape_module.__name__:
            Landscape._landscape_type = cls_type # pylint: disable=protected-access
            break
    return

def startup_landscape(include_ssh: bool=True, include_upnp: bool=True,
                      allow_missing_devices: bool=False, allow_unknown_devices: bool=False,
                      validate_features: bool=True, validate_topology: bool=True,
                      interactive: Optional[bool]=None) -> Landscape:
    """
        Statup the landscape outside of a testrun.
    """

    interactive_mode = False
    if AKIT_VARIABLES.AKIT_INTERACTIVE_CONSOLE:
        interactive_mode = AKIT_VARIABLES.AKIT_INTERACTIVE_CONSOLE

    if interactive is not None:
        interactive_mode = interactive

    # ==================== Landscape Initialization =====================
    # The first stage of standing up the test landscape is to create and
    # initialize the Landscape object.  If more than one thread calls the
    # constructor of the Landscape, object, the other thread will block
    # until the first called has initialized the Landscape and released
    # the gate blocking other callers.

    # When the landscape object is first created, it spins up in configuration
    # mode, which allows consumers consume and query the landscape configuration
    # information.
    lscape = LandscapeSingleton()
    lscape.interactive_mode = interactive_mode

    lscape.activate_configuration()

    from akit.extensionpoints import AKitExtensionPoints
    extension_points = AKitExtensionPoints()

    UpnpCoordinatorIntegrationType = None
    SshPoolCoordinatorIntegrationType = None

    if include_upnp:
        UpnpCoordinatorIntegrationType = extension_points.get_coupling_upnp_coord_integration_type()
    
        # Give the UpnpCoordinatorIntegration an opportunity to register itself, we are
        # doing this in this way to simulate test framework startup.
        UpnpCoordinatorIntegrationType.attach_to_framework(lscape)

    if include_ssh:
        SshPoolCoordinatorIntegrationType = extension_points.get_coupling_ssh_coord_integration_type()

        # Give the SshPoolCoordinatorIntegration an opportunity to register itself, we are
        # doing this in this way to simulate test framework startup.
        SshPoolCoordinatorIntegrationType.attach_to_framework(lscape)

    # After all the coordinators have had an opportunity to register with the
    # 'landscape' object, transition the landscape to the activated 'phase'
    lscape.activate_integration()

    if UpnpCoordinatorIntegrationType is not None:
        # After we transition the the landscape to the activated phase, we give
        # the different coordinators such as the UpnpCoordinatorIntegration an
        # opportunity to attach to its environment and determine if the resources
        # requested and the resource configuration match
        UpnpCoordinatorIntegrationType.attach_to_environment()

    if SshPoolCoordinatorIntegrationType is not None:
        # After we transition the the landscape to the activated phase, we give
        # the different coordinators such as the SshPoolCoordinatorIntegration an
        # opportunity to attach to its environment and determine if the resources
        # requested and the resource configuration match
        SshPoolCoordinatorIntegrationType.attach_to_environment()

    # Finalize the activation process and transition the landscape
    # to fully active where all APIs are available.
    lscape.activate_operations(allow_missing_devices=allow_missing_devices,
                               allow_unknown_devices=allow_unknown_devices,
                               validate_features=validate_features,
                               validate_topology=validate_topology)

    if include_ssh:
        lscape.ssh_coord.establish_presence()

    if include_upnp:
        lscape.upnp_coord.establish_presence()

    return lscape



"""
.. module:: interfaces
    :platform: Darwin, Linux, Unix, Windows
    :synopsis: Contains helper functions for working with internet interfaces

.. moduleauthor:: Myron Walker <myron.walker@gmail.com>
"""

__author__ = "Myron Walker"
__copyright__ = "Copyright 2020, Myron W Walker"
__credits__ = []
__version__ = "1.0.0"
__maintainer__ = "Myron Walker"
__email__ = "myron.walker@gmail.com"
__status__ = "Development" # Prototype, Development or Production
__license__ = "MIT"

from typing import Dict, List, Tuple, Union

from enum import IntEnum

import os
import re
import socket
import subprocess

import netifaces

class InterfaceClass(IntEnum):
    UNKNOWN = 0
    LOOPBACK = 1
    BRIDGE = 2
    WIRED = 3
    WIRELESS = 4
    TUNNEL = 5

def encode_address(address: str) -> bytes:
    """
        Encodes the address string to bytes.

        :param address: The IP address to encode.

        :returns: A packed string suitable for use with low-level network functions.
    """
    is_ipv6 = ':' in address
    address_family = socket.AF_INET6 if is_ipv6 else socket.AF_INET
    return socket.inet_pton(address_family, address)

def get_ipv4_address(ifname: str) -> Union[str, None]:
    """
        Get the first IPv4 address associated with the specified interface name.

        :param ifname: The interface name to lookup the IP address for.

        :returns: The IPv4 address associated with the specified interface name or None
    """
    addr = None

    address_info = netifaces.ifaddresses(ifname)
    if address_info is not None and netifaces.AF_INET in address_info:
        addr_info = address_info[netifaces.AF_INET][0]
        addr = addr_info["addr"]

    return addr

def get_ipv6_address(ifname: str) -> Union[str, None]:
    """
        Get the first IPv6 address associated with the specified interface name.

        :param ifname: The interface name to lookup the IP address for.

        :returns: The IPv6 address associated with the specified interface name or None
    """
    addr = None

    address_info = netifaces.ifaddresses(ifname)
    if address_info is not None and netifaces.AF_INET6 in address_info:
        addr_info = address_info[netifaces.AF_INET][0]
        addr = addr_info["addr"]

    return addr

def get_correspondance_interface(ref_ip: str, ref_port: int, addr_family=socket.AF_INET) -> Tuple[str, str]:
    """
        Utilizes the TCP stack to make a connection to a remote computer and utilizes
        gets the network interface that was used to connect to the remote computer.
        This network interface is the network interface that is likely to be visible
        to the remote computer and thus could be used to establish services that will
        be visible to the remote computer.

        :param ref_ip: An IP address of a computer that is on the subnet that you wish
                       to find the correspondance ip address for and that is hosting a
                       service that will accept a TCP connection from a client.
        :param ref_port: The port number of a service on a computer that will accept a
                         TCP connection so we can determine a path to the computer.
        :param addr_family: The socket address family to utilize when making a remote
                            connection to a host socket.AF_INET or socket.AF_INET6.
                            The address family used will determine the type of IP address
                            returned from this function.

        :returns: The correspondance interface and IPAddress that can be used to setup a
                  service that is visible to the reference IP address.
    """

    corr_iface = None

    corr_ip = get_correspondance_ip_address(ref_ip, ref_port, addr_family=addr_family)

    corr_iface = get_interface_for_ip(corr_ip)

    return corr_iface, corr_ip

def get_correspondance_ip_address(ref_ip: str, ref_port: int, addr_family=socket.AF_INET) -> str:
    """
        Utilizes the TCP stack to make a connection to a remote computer and utilizes
        gets the socket address of the socket that connected to the remote computer.
        This socket address is the address of the socket that is likely to be visible
        to the remote computer and thus could be used to establish services that will
        be visible to the remote computer.

        :param ref_ip: An IP address of a computer that is on the subnet that you wish
                       to find the correspondance ip address for and that is hosting a
                       service that will accept a TCP connection from a client.
        :param ref_port: The port number of a service on a computer that will accept a
                         TCP connection so we can determine a path to the computer.
        :param addr_family: The socket address family to utilize when making a remote
                            connection to a host socket.AF_INET or socket.AF_INET6.
                            The address family used will determine the type of IP address
                            returned from this function.

        :returns: The correspondance IP address that can be used to setup a service that
                  is visible to the reference IP address.
    """
    corr_ip = None

    sock = socket.socket(addr_family, socket.SOCK_STREAM)
    try:
        sock.settimeout(10)
        sock.connect((ref_ip, ref_port))
        corr_ip, _ = sock.getsockname()
    except Exception: # pylint: disable=broad-except
        # If an exception occurs, just return None
        pass
    finally:
        sock.close()

    return corr_ip

def get_interface_for_ip(if_addr: str) -> str:
    """
        Finds the interface name on the local machine for the internet address provided.

        :param if_addr: The internet address on the local machine to find the interface name for.

        :returns: The ifname that corresponds to the address provided.
    """
    addr_info = socket.getaddrinfo(if_addr, 80)
    addr_family=addr_info[0][0]

    if_name = None

    iface_name_list = [ iface for iface in netifaces.interfaces() ]
    for iface in iface_name_list:
        if_address_table = netifaces.ifaddresses(iface)
        if addr_family in if_address_table:
            faddr_list = if_address_table[addr_family]
            for faddr in faddr_list:
                if 'addr' in faddr:
                    ipaddr = faddr['addr']
                    if ipaddr == if_addr:
                        if_name = iface
                        break
        if if_name is not None:
            break

    return if_name

def get_interface_class_table() -> Dict[str, InterfaceClass]:
    """
        Creates a dictionary lookup table of interface names to :class:`InterfaceClass` enumerations.

        :returns: The table of interface names to interface class values.
    """

    results = {}

    iface_name_list = [ iface for iface in netifaces.interfaces() ]
    for ifname in iface_name_list:
        ifcls = get_interface_class(ifname)
        results[ifname] = ifcls

    return results

def get_interface_class(ifname: str) -> InterfaceClass:
    """
        Returns an :class:`InterfaceClass` value for the interface name specified.

        :param ifname: The name of the interface to determine the interface class for.

        :returns: An :class:`InterfaceClass` that corresponds to the specified interface.
    """

    if_syscls_path = "/sys/class/net/{}".format(ifname)

    if not os.path.exists(if_syscls_path):
        raise FileNotFoundError("No sys class path found for interface={} exp={}".format(ifname, if_syscls_path))
    
    if_files_list = os.listdir(if_syscls_path)
    
    if_class = InterfaceClass.UNKNOWN
    if "wireless" in if_files_list:
        if_class = InterfaceClass.WIRELESS
    elif "bridge" in if_files_list:
        if_class = InterfaceClass.BRIDGE
    else:
        ifinfo = subprocess.check_output("ifconfig {} | grep flags".format(ifname), shell=True).decode('utf8')
        mobj: re.Match = re.match(r"[^<]*.([^>]*)", ifinfo)
        if mobj is not None:
            ifflags = mobj.groups()[0].strip().split(",")
            if "LOOPBACK" in ifflags:
                if_class = InterfaceClass.LOOPBACK
            elif "POINTTOPOINT" in ifflags:
                if_class = InterfaceClass.TUNNEL
            else:
                if_class = InterfaceClass.WIRED
    
    return if_class

def get_interface_names_of_class(ifcls_filter: InterfaceClass) -> List[str]:
    """
        Gets a list of interface names of a give :class:`InterfaceClass`.

        :returns: List of filtered interface names.
    """
    
    iface_list = []

    for ifname, ifcls in get_interface_class_table().items():
        if ifcls == ifcls_filter:
            iface_list.append(ifname)

    return iface_list

def is_ipv6_address(candidate: str) -> bool:
    """
        Checks to see if 'candidate' is an ipv6 address.

        :param candidate: A string that is to be checked to see if it is a valid IPv6 address.

        :returns: A boolean indicating if an IP address is an IPv6 address
    """
    is_ipv6 = False
    if len(candidate) == 16:
        is_ipv6 = True

    return is_ipv6



"""
.. module:: foundations
    :platform: Darwin, Linux, Unix, Windows
    :synopsis: Module which contains framework logging functions which extend the functionality to
        the python :module:`logging` module.

.. note:: The modules that are named `xsomething` like this module are prefixed with an `x` character to
        indicate they extend the functionality of a base python module and the `x` is pre-pended to
        prevent module name collisions with python modules.

.. moduleauthor:: Myron Walker <myron.walker@gmail.com>
"""

__author__ = "Myron Walker"
__copyright__ = "Copyright 2020, Myron W Walker"
__credits__ = []
__version__ = "1.0.0"
__maintainer__ = "Myron Walker"
__email__ = "myron.walker@gmail.com"
__status__ = "Development" # Prototype, Development or Production
__license__ = "MIT"


import fnmatch
import logging
import os
import sys
import time
import traceback


from akit.environment.context import Context
from akit.environment.variables import AKIT_VARIABLES


class AKitLogLevels:
    NOTSET = logging.NOTSET
    DEBUG = logging.DEBUG
    INFO = logging.INFO
    WARNING = logging.WARNING
    ERROR = logging.ERROR
    CRITICAL = logging.CRITICAL
    SECTION = 100
    RENDER = 200
    QUIET = 1000


logging.addLevelName(AKitLogLevels.SECTION, "SECTION")
logging.addLevelName(AKitLogLevels.RENDER, "RENDER")
logging.addLevelName(AKitLogLevels.QUIET, "QUIET")


# Start Logging to Standard Out.  We need to make sure it is initialized to something as early as possible,
# but we may not have a file to write to yet until logging_initialize is called by a proper entry point
logging.basicConfig(level=logging.NOTSET)


DEFAULT_LOGGER_NAME = "AKIT"

LOGGING_SECTION_MARKER = "="
LOGGING_SECTION_MARKER_LENGTH = 80

DEFAULT_LOGFILE_FORMAT = '[%(asctime)s][%(name)s][%(levelname)s][%(message)s]'


def format_log_section_header(title):
    """
        Formats a log section header by centering the title inside of the section marker character string.
    """
    title_upper = " %s " % title.strip().upper()
    marker_count = LOGGING_SECTION_MARKER_LENGTH - len(title_upper)
    marker_half = marker_count >> 1
    marker_prefix = LOGGING_SECTION_MARKER * marker_half
    marker_suffix = LOGGING_SECTION_MARKER * (marker_count - marker_half)
    header = "\n%s%s%s\n" % (marker_prefix, title_upper, marker_suffix)
    return header


OTHER_LOGGER_FILTERS = []


class AKitLogFormatter(logging.Formatter):
    """
    Formatter instances are used to convert a LogRecord to text.

    Formatters need to know how a LogRecord is constructed. They are
    responsible for converting a LogRecord to (usually) a string which can
    be interpreted by either a human or an external system. The base Formatter
    allows a formatting string to be specified. If none is supplied, the
    style-dependent default value, "%(message)s", "{message}", or
    "${message}", is used.

    The Formatter can be initialized with a format string which makes use of
    knowledge of the LogRecord attributes - e.g. the default value mentioned
    above makes use of the fact that the user's message and arguments are pre-
    formatted into a LogRecord's message attribute. Currently, the useful
    attributes in a LogRecord are described by:

    %(name)s            Name of the logger (logging channel)
    %(levelno)s         Numeric logging level for the message (DEBUG, INFO,
                        WARNING, ERROR, CRITICAL)
    %(levelname)s       Text logging level for the message ("DEBUG", "INFO",
                        "WARNING", "ERROR", "CRITICAL")
    %(pathname)s        Full pathname of the source file where the logging
                        call was issued (if available)
    %(filename)s        Filename portion of pathname
    %(module)s          Module (name portion of filename)
    %(lineno)d          Source line number where the logging call was issued
                        (if available)
    %(funcName)s        Function name
    %(created)f         Time when the LogRecord was created (time.time()
                        return value)
    %(asctime)s         Textual time when the LogRecord was created
    %(msecs)d           Millisecond portion of the creation time
    %(relativeCreated)d Time in milliseconds when the LogRecord was created,
                        relative to the time the logging module was loaded
                        (typically at application startup time)
    %(thread)d          Thread ID (if available)
    %(threadName)s      Thread name (if available)
    %(process)d         Process ID (if available)
    %(message)s         The result of record.getMessage(), computed just as
                        the record is emitted
    """

    converter = time.localtime

    def __init__(self, fmt=None, datefmt=None, style='%', validate=True):
        """
        Initialize the formatter with specified format strings.

        Initialize the formatter either with the specified format string, or a
        default as described above. Allow for specialized date formatting with
        the optional datefmt argument. If datefmt is omitted, you get an
        ISO8601-like (or RFC 3339-like) format.

        Use a style parameter of '%', '{' or '$' to specify that you want to
        use one of %-formatting, :meth:`str.format` (``{}``) formatting or
        :class:`string.Template` formatting in your format string.

        .. versionchanged:: 3.2
           Added the ``style`` parameter.
        """
        super().__init__(fmt=fmt, datefmt=datefmt, style=style, validate=validate)
        return

    def formatException(self, ei):
        """
        Format and return the specified exception information as a string.

        This default implementation just uses
        traceback.print_exception()
        """
        s = super().formatException(ei)
        return s

    def formatStack(self, stack_info):
        """
        This method is provided as an extension point for specialized
        formatting of stack information.

        The input data is a string as returned from a call to
        :func:`traceback.print_stack`, but with the last trailing newline
        removed.

        The base implementation just returns the value passed in.
        """
        return stack_info

    def format(self, record):
        """
        Format the specified record as text.

        The record's attribute dictionary is used as the operand to a
        string formatting operation which yields the returned string.
        Before formatting the dictionary, a couple of preparatory steps
        are carried out. The message attribute of the record is computed
        using LogRecord.getMessage(). If the formatting string uses the
        time (as determined by a call to usesTime(), formatTime() is
        called to format the event time. If there is exception information,
        it is formatted using formatException() and appended to the message.
        """
        record.message = record.getMessage()

        if record.levelno < AKitLogLevels.SECTION:
            s = super().format(record)
        else:
            s = record.message

        return s


class LoggingManagerWrapper:
    """
        There is [under normal circumstances] just one Manager instance, which
        holds the hierarchy of loggers.
    """
    # pylint: disable=protected-access

    def __init__(self, manager):
        """
            Initialize the manager with the root node of the logger hierarchy.
        """
        self.manager = manager
        return

    @property
    def disable(self):
        """
            Git if logging is disabled.
        """
        return self.manager.disable

    @property
    def emittedNoHandlerWarning(self):
        """
            Get if the emittion of handler warnings is turned off.
        """
        return self.manager.emittedNoHandlerWarning

    @property
    def loggerClass(self):
        """
            Returns the logger Class.
        """
        return self.manager.loggerClass

    @property
    def loggerDict(self):
        """
            Returns a dictionary of the loggers.
        """
        return self.manager.loggerDict

    @property
    def logRecordFactory(self):
        """
            Returns the log record factory.
        """
        return self.manager.logRecordFactory

    @property
    def root(self):
        """
            Returns the root logger.
        """
        return self.manager.root

    def getLogger(self, name):
        """
            Get a logger with the specified name (channel name), creating it
            if it doesn't yet exist. This name is a dot-separated hierarchical
            name, such as "a", "a.b", "a.b.c" or similar.

            If a PlaceHolder existed for the specified name [i.e. the logger
            didn't exist but a child of it did], replace it with the created
            logger and fix up the parent/child references which pointed to the
            placeholder to now point to the logger.
        """
        rv = None

        if not isinstance(name, str):
            raise TypeError('A logger name must be a string')

        logging._acquireLock()
        try:
            if name in self.loggerDict:
                rv = self.loggerDict[name]
                if isinstance(rv, logging.PlaceHolder):
                    ph = rv
                    rv = (self.loggerClass or logging._loggerClass)(name)
                    rv.manager = self
                    self.loggerDict[name] = rv
                    self._fixupChildren(ph, rv)
                    self._fixupParents(rv)
            else:
                rv = (self.loggerClass or logging._loggerClass)(name)
                rv.manager = self
                self.loggerDict[name] = rv
                self._fixupParents(rv)
        finally:
            logging._releaseLock()

        return rv

    def matchLoggers(self, pattern):
        """
            Finds all the loggers whose name match the associated pattern.
        """
        logger_dict = {}

        logging._acquireLock()
        try:
            for lname in self.manager.loggerDict.keys():
                if fnmatch.fnmatch(lname, pattern):
                    logger_dict[lname] = self.manager.loggerDict[lname]
        finally:
            logging._releaseLock()

        return logger_dict

    def setLoggerClass(self, klass):
        """
            Set the class to be used when instantiating a logger with this Manager.
        """
        self.manager.setLoggerClass(klass)
        return

    def setLogRecordFactory(self, factory):
        """
            Set the factory to be used when instantiating a log record with this
            Manager.
        """
        self.manager.setLogRecordFactory(factory)
        return

    def _fixupParents(self, alogger):
        """
            Ensure that there are either loggers or placeholders all the way
            from the specified logger to the root of the logger hierarchy.
        """
        self.manager._fixupParents(alogger)
        return

    def _fixupChildren(self, ph, alogger):
        """
            Ensure that children of the placeholder ph are connected to the
            specified logger.
        """
        self.manager._fixupChildren(ph, alogger)
        return

    def _clear_cache(self):
        """
            Clear the cache for all loggers in loggerDict
            Called when level changes are made
        """
        self.manager._clear_cache()
        return


class LoggingDefaults:
    """
        Makes all the default values associated with logging available.
    """
    DefaultFileLoggingHandler = logging.FileHandler



class AKitLoggerWrapper:
    """
        We utilize a log wrapper so we can re-initialize logging and switch out the logger
        without invalidating references to the logger that we have given out.
    """

    def __init__(self, logger):
        self._logger = logger
        return

    def critical(self, msg, *args, **kwargs):
        """
            Log 'msg % args' with severity 'CRITICAL'.

            To pass exception information, use the keyword argument exc_info with
            a true value, e.g.

            logger.critical("Houston, we have a %s", "major disaster", exc_info=1)
        """
        self._logger.critical(msg, *args, **kwargs)
        return

    def debug(self, msg, *args, **kwargs):
        """
            Log 'msg % args' with severity 'DEBUG'.

            To pass exception information, use the keyword argument exc_info with
            a true value, e.g.

            logger.debug("Houston, we have a %s", "thorny problem", exc_info=1)
        """
        self._logger.debug(msg, *args, **kwargs)
        return

    def test_begin(self, testname, **test_args):
        """
        """
        info_msg_lines = [
            "TEST BEGIN - {}".format(testname),
            "    ARGS:"
        ]

        for arg_name, arg_value in test_args.items():

            arg_value_debug = None
            if hasattr(arg_value, "__debug_repr__"):
                arg_value_debug= arg_value.__debug_repr__()
            else:
                arg_value_debug = str(arg_value)

            info_msg_lines.append("    {} = {}".format(arg_name, arg_value_debug))

        info_msg = os.linesep.join(info_msg_lines)

        self._logger.info(info_msg)
        return

    def test_end(self, testname):
        self._logger.info("TEST END - {}".format(testname))
        return

    def error(self, msg, *args, **kwargs):
        """
            Log 'msg % args' with severity 'ERROR'.

            To pass exception information, use the keyword argument exc_info with
            a true value, e.g.

            logger.error("Houston, we have a %s", "major problem", exc_info=1)
        """
        self._logger.error(msg, *args, **kwargs)
        return

    def exception(self, msg, *args, exc_info=True, **kwargs):
        """
            Convenience method for logging an ERROR with exception information.
        """
        self._logger.exception(msg, *args, exc_info=exc_info, **kwargs)
        return

    def fatal(self, msg, *args, **kwargs):
        """
            Log 'msg % args' with severity 'CRITICAL'.

            To pass exception information, use the keyword argument exc_info with
            a true value, e.g.

            logger.critical("Houston, we have a %s", "major disaster", exc_info=1)
        """
        self._logger.critical(msg, *args, **kwargs)
        return

    def info(self, msg, *args, **kwargs):
        """
            Log 'msg % args' with severity 'INFO'.

            To pass exception information, use the keyword argument exc_info with
            a true value, e.g.

            logger.info("Houston, we have a %s", "interesting problem", exc_info=1)
        """
        self._logger.info(msg, *args, **kwargs)
        return

    def render(self, line):
        """
            Logs a log section marker
        """
        self._logger.log(AKitLogLevels.RENDER, line)
        return

    def section(self, title):
        """
            Logs a log section marker
        """
        marker = format_log_section_header(title)
        self._logger.log(AKitLogLevels.SECTION, marker)
        return

    def warning(self, msg, *args, **kwargs):
        """
            Log 'msg % args' with severity 'WARNING'.

            To pass exception information, use the keyword argument exc_info with
            a true value, e.g.

            logger.warning("Houston, we have a %s", "bit of a problem", exc_info=1)
        """
        self._logger.warning(msg, *args, **kwargs)
        return

    def warn(self, msg, *args, **kwargs):
        """
            Log 'msg % args' with severity 'WARNING'.

            To pass exception information, use the keyword argument exc_info with
            a true value, e.g.

            logger.warning("Houston, we have a %s", "bit of a problem", exc_info=1)
        """
        self._logger.warning(msg, *args, **kwargs)
        return


class LessThanRecordFilter(logging.Filter):
    """
        Filters records with a log level < WARNING
    """

    def __init__(self, filter_at_level):
        super(LessThanRecordFilter, self).__init__("LessThanFilter")
        self._filter_at_level = filter_at_level
        return

    def filter(self, record): # pylint: disable=no-self-use
        """
            Performs the filtering of records.
        """
        process_rec = record.levelno < self._filter_at_level
        return process_rec


class OtherFilter(logging.Filter):
    """
        Filters records with a name that match a prefix expression and marks them as other.
    """
    def __init__(self, prefix):
        super(OtherFilter, self).__init__("OtherFilter")
        self.prefix = prefix
        return

    def filter(self, record):
        """
            Performs the filtering of records.
        """
        record.is_other = fnmatch.fnmatch(record.name, self.prefix)
        return record.is_other


class RelevantFilter(logging.Filter):
    """
        Allows records that are not marked as other.
    """
    def __init__(self):
        super(RelevantFilter, self).__init__("RelevantFilter")
        return

    def filter(self, record): # pylint: disable=no-self-use
        """
            Performs the filtering of records.
        """
        process_rec = True
        if hasattr(record, "is_other"):
            if record.is_other:
                process_rec = False
        return process_rec


logging_initialized = False


def logging_initialize():
    """
        Method used to configure the automation kit logging based on the environmental parameters
        specified and then reinitialize the logging.
    """
    global logging_initialized

    if not logging_initialized:
        logging_initialized = True

        ctx = Context()

        env = ctx.lookup("/environment")
        conf = ctx.lookup("/configuration")
        logging_conf = conf["logging"]

        consolelevel = AKIT_VARIABLES.AKIT_LOG_LEVEL_CONSOLE
        logfilelevel = AKIT_VARIABLES.AKIT_LOG_LEVEL_FILE

        logname_template = logging_conf["logname"]
        logname = env.fill_template(logname_template)

        output_directory = env["output_directory"]
        if not os.path.exists(output_directory):
            os.makedirs(output_directory)

        log_branches = []
        if "branched" in logging_conf:
            log_branches = logging_conf["branched"]


        # Setup the log files
        _reinitialize_logging(consolelevel, logfilelevel, output_directory, logname, log_branches)

    return


def logging_create_branch_logger(logger_name, logfilename, log_level):
    """
        Method that allows for the creation of a separate logfile for specific loggers in order
        to reduce the noise in the main logfile.  A common use for this would be to redirect
        logging from specific modules such as 'paramiko' and 'httplib' to thier own log files.

        :param logger_name: The name of the logger to create a branch log for.
    """
    root_logger = logging.getLogger()

    if logger_name in root_logger.manager.loggerDict:
        target_logger = logging.getLogger(logger_name)

        # Setup the relevant log file which will get all the
        # log entries from loggers that satisified a relevant
        # logger name prefix match
        handler = logging.FileHandler(logfilename)
        handler.setLevel(log_level)
        for handler in target_logger.handlers:
            target_logger.removeHandler(handler)
        target_logger.addHandler(handler)

    return


def _reinitialize_logging(consolelevel, logfilelevel, output_dir, logfile_basename, log_branches):
    """
        Helper method to re-initialize the logging when the path to the output directory changes
        shortly after startup of the framework.  This method also handles the configuration of
        output levels, stdout and stderr file wrappers.
    """

    consolelevel_strerr = logging.WARNING

    console_filter = LessThanRecordFilter(logging.WARNING)
    console_filter_stderr = None

    if isinstance(consolelevel, str):
        consolelevel_upper = consolelevel.upper()
        if consolelevel_upper == "QUIET":
            consolelevel = AKitLogLevels.QUIET
            consolelevel_strerr = AKitLogLevels.QUIET
            console_filter = LessThanRecordFilter(AKitLogLevels.QUIET)
            console_filter_stderr = LessThanRecordFilter(AKitLogLevels.QUIET)

        elif hasattr(logging, consolelevel_upper):
            consolelevel = getattr(logging, consolelevel_upper)
        else:
            consolelevel = logging.INFO
    else:
        print("")
        print("NOTE: Console logging set to %r" % consolelevel)
        print("NOTE: outputdir=%s" % output_dir)
        print("")

        consolelevel = logging.WARNING

    basecomp, extcomp = os.path.splitext(logfile_basename)

    ctx = Context()
    env = ctx.lookup("/environment")

    debug_logfilename = os.path.join(output_dir, basecomp + ".DEBUG" + extcomp)
    other_logfilename = os.path.join(output_dir, basecomp + ".OTHER" + extcomp)

    env["logfile_debug"] = debug_logfilename
    env["logfile_other"] = other_logfilename

    rel_logfilename = os.path.join(output_dir, basecomp + extcomp)

    logging.Logger.manager = LoggingManagerWrapper(logging.Logger.manager)

    # Remove all the log handlers from the root logger
    root_logger = logging.getLogger()
    for lhandler in root_logger.handlers:
        root_logger.removeHandler(lhandler)

    # Set the root logger to NOTSET, so we don't impose an effective log
    # level on child loggers
    root_logger.setLevel(logging.NOTSET)

    # Setup the debug logfile
    base_handler = LoggingDefaults.DefaultFileLoggingHandler(debug_logfilename)
    base_handler.setFormatter(AKitLogFormatter(DEFAULT_LOGFILE_FORMAT))
    base_handler.setLevel(logging.NOTSET)
    root_logger.addHandler(base_handler)

    # Setup the other log handler and other filter, we
    # need to add the other log handler before adding
    # the relevant log handler
    other_handler = LoggingDefaults.DefaultFileLoggingHandler(other_logfilename)
    other_handler.setFormatter(AKitLogFormatter(DEFAULT_LOGFILE_FORMAT))
    other_handler.setLevel(logfilelevel)
    for other_expr in OTHER_LOGGER_FILTERS:
        other_handler.addFilter(OtherFilter(other_expr))
    root_logger.addHandler(other_handler)

    # Setup the relevant log file which will get all the
    # log entries from loggers that satisified a relevant
    # logger name prefix match
    rel_handler = LoggingDefaults.DefaultFileLoggingHandler(rel_logfilename)
    rel_handler.setFormatter(AKitLogFormatter(DEFAULT_LOGFILE_FORMAT))
    rel_handler.setLevel(logfilelevel)
    rel_handler.addFilter(RelevantFilter())
    root_logger.addHandler(rel_handler)

    # Setup the stdout logger with the correct console level and
    # filter the log entries from the stdout handler that are
    # greater than Info level
    stdout_logger = logging.StreamHandler(sys.stdout)
    stdout_logger.setLevel(consolelevel)
    stdout_logger.addFilter(console_filter)

    stderr_logger = logging.StreamHandler(sys.stderr)
    stderr_logger.setLevel(consolelevel_strerr)
    if console_filter_stderr is not None:
        stderr_logger.addFilter(console_filter_stderr)

    root_logger.addHandler(stdout_logger)
    root_logger.addHandler(stderr_logger)

    for binfo in log_branches:
        try:
            logger_name = binfo["name"]
            logfilename = binfo["logname"]
            log_level = binfo["loglevel"]

            logging_create_branch_logger(logger_name, logfilename, log_level)
        except Exception: # pylint: disable=broad-except
            errmsg = "Error configuration branch logger." + os.linesep
            errmsg = traceback.format_exc()
            root_logger.error(errmsg)

    akit_logger = getAutomatonKitLogger()
    akit_logger.section("Logging Initialized")

    return


akit_logger_table = {}


def getAutomatonKitLogger(logger_name: str=None) -> AKitLoggerWrapper:
    """
        Gets an automation kit logger by name. AutomationKit loggers are different
        in that they are a logger wrapper which means the characteristics of the loggers
        can be changed latter without corrupting any previous references handed out to
        any particular loggers.

        :param logger_name: The name of the logger to obtain. The default is the AutomationKit logger.
    """
    global akit_logger_table

    if logger_name is None:
        logger_name = DEFAULT_LOGGER_NAME

    logger = None

    if logger_name in akit_logger_table:
        logger = akit_logger_table[logger_name]
    else:
        logger = AKitLoggerWrapper(logging.getLogger(logger_name))

    return logger



"""
.. module:: looperpool
    :platform: Darwin, Linux, Unix, Windows
    :synopsis: Module that contains a Looper for repeating a loop function useful
        for repeated, work packet or queue processing.

.. moduleauthor:: Myron Walker <myron.walker@gmail.com>
"""

__author__ = "Myron Walker"
__copyright__ = "Copyright 2020, Myron W Walker"
__credits__ = []
__version__ = "1.0.0"
__maintainer__ = "Myron Walker"
__email__ = "myron.walker@gmail.com"
__status__ = "Development" # Prototype, Development or Production
__license__ = "MIT"

from threading import RLock

from akit.exceptions import AKitLooperError

from akit.xthreading.looper import Looper
from akit.xthreading.looperqueue import LooperQueue, LooperQueueShutdown

class LooperPool:
    """
        The :class:`LooperPool` provides a convenient way to setup a thread pool and worker threads.  The
        :class:`LooperPool` is passed a type derived from :class:`Looper` in order to customize the work
        performed by the :class:`LooperPool`
    """

    def __init__(self, looper_type: Looper, group_name: str=None, min_loopers: int=5, max_loopers: int=10, highwater: int=5, daemon=True, **kwargs):
        self._looper_type = looper_type
        self._group_name = group_name
        self._min_loopers = min_loopers
        self._max_loopers = max_loopers
        self._highwater = highwater
        self._daemon = daemon
        self._kwargs = kwargs

        self._queue = LooperQueue()

        self._running = False

        self._threads_lock = RLock()
        self._threads = []
        self._thread_count = 0
        return

    def push_work(self, packet: object):
        """
            Pushes a work packet for the :class:`LooperPool` threads to work on.
        """

        if self._running:
            available = self._queue.push_work(packet)
            if available > self._highwater:
                self._threads_lock.acquire()
                try:
                    if self._thread_count < self._max_loopers:
                        self._locked_start_looper()
                finally:
                    self._threads_lock.release()
        else:
            raise AKitLooperError("LooperPool: push_work called after the looper pool has been shutdown.") from None

        return available

    def push_work_packets(self, packets: list):
        """
            Pushes a list of work packets for the :class:`LooperPool` threads to work on.
        """

        if self._running:
            available = self._queue.push_work_packets(packets)
            if available > self._highwater:
                self._threads_lock.acquire()
                try:
                    if self._thread_count < self._max_loopers:
                        self._locked_start_looper()
                finally:
                    self._threads_lock.release()
        else:
            raise AKitLooperError("LooperPool: push_work_packets called after the looper pool has been shutdown.") from None

        return available

    def shutdown(self):
        """
            Shutdown the :class:`LooperPool` and its worker threads.
        """

        # Set running to False to disallow the queueing of new work
        self._running = False

        running_threads = None
        self._threads_lock.acquire()
        try:
            running_threads = [th for th in self._threads]
            self._running = False
            shutdown_work = [ LooperQueueShutdown() for n in range(0, len(running_threads) * 2)]
            self._threads_lock.release()
            try:
                self._queue.push_work_packets(shutdown_work)
            finally:
                self._threads_lock.acquire()
        finally:
            self._threads_lock.release()

        for th in running_threads:
            th.join()

        return

    def start_pool(self):
        """
            Starts the :class:`LooperPool` with the minimum number of threads.
        """

        if self._running:
            raise AKitLooperError("LooperPool: start called while LooperPool is already running") from None

        self._threads_lock.acquire()
        self._running = True
        try:
            # We start up the minimum number of threads
            for _ in range(0, self._min_loopers):
                self._locked_start_looper()
        finally:
            self._threads_lock.release()

        return

    def _locked_start_looper(self):
        """
            Starts up a new :class:`Looper` worker thread.
        """

        try:
            self._thread_count += 1
            looper_name = "%s-%d" % (self._group_name, self._thread_count)
            looper = self._looper_type(self._queue, name=looper_name, group=self._group_name, daemon=self._daemon, **self._kwargs)
            looper.start()
            self._threads.append(looper)
        except:
            self._thread_count -= 1
            raise

        return



"""

    NOTE: This is a code generated file.  This file should not be edited directly.
"""



from akit.aspects import AspectsUPnP, DEFAULT_UPNP_ASPECTS

from akit.extensible import LoadableExtension
from akit.interop.upnp.services.upnpserviceproxy import UpnpServiceProxy

class HVAC_FanOperatingMode1ServiceProxy(UpnpServiceProxy, LoadableExtension):
    """
        This is a code generated proxy class to the 'urn:schemas-upnp-org:service:HVAC_FanOperatingMode:1' service.
    """

    SERVICE_MANUFACTURER = 'UPnP'
    SERVICE_TYPE = 'HVAC_FanOperatingMode1'

    SERVICE_DEFAULT_VARIABLES = {
        "FanStatus": { "data_type": "string", "default": "On", "allowed_list": "['On', 'Off']"},
        "Mode": { "data_type": "string", "default": "Auto", "allowed_list": "['Auto', 'ContinuousOn', 'PeriodicOn']"},
        "Name": { "data_type": "string", "default": None, "allowed_list": None},
    }

    SERVICE_EVENT_VARIABLES = {}

    def action_GetFanStatus(self, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetFanStatus action.

            :returns: "CurrentStatus"
        """
        arguments = { }

        out_params = self.call_action("GetFanStatus", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("CurrentStatus",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetMode(self, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetMode action.

            :returns: "CurrentMode"
        """
        arguments = { }

        out_params = self.call_action("GetMode", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("CurrentMode",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetName(self, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetName action.

            :returns: "CurrentName"
        """
        arguments = { }

        out_params = self.call_action("GetName", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("CurrentName",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_SetMode(self, NewMode, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the SetMode action.
        """
        arguments = {
            "NewMode": NewMode,
        }

        self.call_action("SetMode", arguments=arguments, aspects=aspects)

        return

    def action_SetName(self, NewName, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the SetName action.
        """
        arguments = {
            "NewName": NewName,
        }

        self.call_action("SetName", arguments=arguments, aspects=aspects)

        return



__author__ = "Myron Walker"
__copyright__ = "Copyright 2020, Myron W Walker"
__credits__ = []
__version__ = "1.0.0"
__maintainer__ = "Myron Walker"
__email__ = "myron.walker@gmail.com"
__status__ = "Development" # Prototype, Development or Production
__license__ = "MIT"

from typing import List

import os

from datetime import datetime

from akit.environment.context import Context
from akit.environment.contextpaths import ContextPaths
from akit.environment.variables import AKIT_VARIABLES
from akit.environment.configuration import OVERRIDE_CONFIGURATION, load_runtime_configuration

ctx = Context()

def override_build_branch(branch_name: str):
    """
        This override function provides a mechanism overriding the AKIT_BUILD_BRANCH
        variable and context configuration setting.

        :param branch: The name of the branch the build came from.
    """
    ctx.insert(ContextPaths.BUILD_BRANCH, branch_name)
    AKIT_VARIABLES.AKIT_BUILD_BRANCH = branch_name
    return

def override_build_flavor(build_flavor: str):
    """
        This override function provides a mechanism overriding the AKIT_BUILD_FLAVOR
        variable and context configuration setting.

        :param build_flavor: The flavor of the build associated with a job.
    """
    ctx.insert(ContextPaths.BUILD_FLAVOR, build_flavor)
    AKIT_VARIABLES.AKIT_BUILD_FLAVOR = build_flavor
    return

def override_build_name(build_name: str):
    """
        This override function provides a mechanism overriding the AKIT_BUILD_NAME
        variable and context configuration setting.

        :param build_name: The build version of the build.
    """
    ctx.insert(ContextPaths.BUILD_NAME, build_name)
    AKIT_VARIABLES.AKIT_BUILD_NAME = build_name
    return

def override_build_url(build_url: str):
    """
        This override function provides a mechanism overriding the AKIT_BUILD_URL
        variable and context configuration setting.

        :param build_url: The url associated with the build.
    """
    ctx.insert(ContextPaths.BUILD_URL, build_url)
    AKIT_VARIABLES.AKIT_BUILD_URL = build_url
    return

def override_config_credentials(filename: str):
    """
        This override function provides a mechanism overriding the AKIT_CONFIG_CREDENTIALS
        variable and context configuration setting.

        :param filename: The full path to the file to set as the credentials file.
    """
    ctx.insert(ContextPaths.CONFIG_FILE_CREDENTIALS, filename)
    AKIT_VARIABLES.AKIT_CONFIG_CREDENTIALS = filename
    return

def override_config_landscape(filename: str):
    """
        This override function provides a mechanism overriding the AKIT_CONFIG_LANDSCAPE
        variable and context configuration setting.

        :param filename: The full path to the file to set as the landscape file.
    """
    ctx.insert(ContextPaths.CONFIG_FILE_LANDSCAPE, filename)
    AKIT_VARIABLES.AKIT_CONFIG_LANDSCAPE = filename
    return

def override_config_landscape_name(landscape_name: str):
    """
        This override function provides a mechanism overriding the AKIT_CONFIG_LANDSCAPE_NAME
        variable and context configuration setting.

        :param landscape_name: The name of the landscape to use when selecting a landscape file.
    """
    last_landscape_name = AKIT_VARIABLES.AKIT_CONFIG_LANDSCAPE_NAME

    ctx.insert(ContextPaths.CONFIG_FILE_LANDSCAPE_NAME, landscape_name)
    AKIT_VARIABLES.AKIT_CONFIG_LANDSCAPE_NAME = landscape_name

    rf_head, rf_middle, rf_tail = AKIT_VARIABLES.AKIT_CONFIG_LANDSCAPE.rpartition(last_landscape_name, landscape_name)
    if rf_middle != landscape_name:
        errmsg = "We should have found the name '{}' in the landscape file path.".format(last_landscape_name)
        raise RuntimeError(errmsg)

    landscape_filename = "".join(rf_head, landscape_name, rf_tail)
    ctx.insert(ContextPaths.CONFIG_FILE_LANDSCAPE, landscape_filename)
    AKIT_VARIABLES.AKIT_CONFIG_LANDSCAPE = landscape_filename
    return

def override_config_runtime(filename: str):
    """
        This override function provides a mechanism overriding the AKIT_CONFIG_RUNTIME
        variable and context configuration setting.

        :param filename: The full path to the file to set as the runtime file.
    """
    ctx.insert(ContextPaths.CONFIG_FILE_RUNTIME, filename)
    AKIT_VARIABLES.AKIT_CONFIG_RUNTIME = filename
    return

def override_config_runtime_name(runtime_name: str):
    """
        This override function provides a mechanism overriding the AKIT_CONFIG_RUNTIME_NAME
        variable and context configuration setting.

        :param runtime_name: The name of the runtime to use when selecting a runtime file.
    """
    last_runtime_name = AKIT_VARIABLES.AKIT_CONFIG_RUNTIME_NAME

    ctx.insert(ContextPaths.CONFIG_FILE_RUNTIME_NAME, runtime_name)
    AKIT_VARIABLES.AKIT_CONFIG_RUNTIME_NAME = runtime_name

    if AKIT_VARIABLES.AKIT_CONFIG_RUNTIME_SEARCH_PATH is None:
        rf_head, rf_middle, rf_tail = AKIT_VARIABLES.AKIT_CONFIG_RUNTIME.rpartition(last_runtime_name)
        if rf_middle != last_runtime_name:
            errmsg = "We should have found the name '{}' in the runtime file path.".format(last_runtime_name)
            raise RuntimeError(errmsg)

        runtime_filename = "".join(rf_head, runtime_name, rf_tail)
        ctx.insert(ContextPaths.CONFIG_FILE_RUNTIME, runtime_filename)
        AKIT_VARIABLES.AKIT_CONFIG_RUNTIME = runtime_filename
    else:
        runtime_search_path_list = AKIT_VARIABLES.AKIT_CONFIG_RUNTIME_SEARCH_PATH.split(":")

        found_runtime_file = None
        for spath in runtime_search_path_list:
            check_path = os.path.join(spath, runtime_name + ".yaml")
            if os.path.exists(check_path):
                found_runtime_file = check_path
                break
        
        ctx.insert(ContextPaths.CONFIG_FILE_RUNTIME, found_runtime_file)
        AKIT_VARIABLES.AKIT_CONFIG_RUNTIME = found_runtime_file
    
    runtime_config = load_runtime_configuration()
    OVERRIDE_CONFIGURATION.update(runtime_config)

    return

def override_config_runtime_search_path(runtime_search_path: str):
    """
        This override function provides a mechanism overriding the AKIT_CONFIG_RUNTIME_NAME
        variable and context configuration setting.

        :param runtime_search_path: The search path to use when looking for the runtime file.
    """
    last_runtime_name = AKIT_VARIABLES.AKIT_CONFIG_RUNTIME_NAME
    runtime_search_path_list = runtime_search_path.split(":")

    found_runtime_file = None
    for spath in runtime_search_path_list:
        check_path = os.path.join(spath, last_runtime_name + ".yaml")
        if os.path.exists(check_path):
            found_runtime_file = check_path
            break

    ctx.insert(ContextPaths.CONFIG_SEARCH_PATH_RUNTIME, runtime_search_path)
    AKIT_VARIABLES.AKIT_CONFIG_RUNTIME_SEARCH_PATH = runtime_search_path

    ctx.insert(ContextPaths.CONFIG_FILE_RUNTIME, found_runtime_file)
    AKIT_VARIABLES.AKIT_CONFIG_RUNTIME = found_runtime_file

    return

def override_config_topology(filename: str):
    """
        This override function provides a mechanism overriding the AKIT_CONFIG_TOPOLOGY
        variable and context configuration setting.

        :param filename: The full path to the file to set as the topology file.
    """
    ctx.insert(ContextPaths.CONFIG_FILE_TOPOLOGY, filename)
    AKIT_VARIABLES.AKIT_CONFIG_TOPOLOGY = filename
    return

def override_config_topology_name(topology_name: str):
    """
        This override function provides a mechanism overriding the AKIT_CONFIG_TOPOLOGY_NAME
        variable and context configuration setting.

        :param topology_name: The name of the topology to use when selecting a topology file.
    """
    last_topology_name = AKIT_VARIABLES.AKIT_CONFIG_TOPOLOGY_NAME

    ctx.insert(ContextPaths.CONFIG_FILE_TOPOLOGY_NAME, topology_name)
    AKIT_VARIABLES.AKIT_CONFIG_TOPOLOGY_NAME = topology_name

    tf_head, tf_middle, tf_tail = AKIT_VARIABLES.AKIT_CONFIG_TOPOLOGY.rpartition(last_topology_name, topology_name)
    if tf_middle != last_topology_name:
        errmsg = "We should have found the name '{}' in the topology file path.".format(last_topology_name)
        raise RuntimeError(errmsg)

    topology_filename = "".join(tf_head, topology_name, tf_tail)
    ctx.insert(ContextPaths.CONFIG_FILE_TOPOLOGY_NAME, topology_filename)
    AKIT_VARIABLES.AKIT_CONFIG_TOPOLOGY_NAME = topology_filename

    return

def override_config_user(filename: str):
    """
        This override function provides a mechanism overriding the AKIT_CONFIG_USER
        variable and context configuration setting.

        :param filename: The full path to the file to set as the user file.
    """
    ctx.insert(ContextPaths.CONFIG_FILE_USER, filename)
    AKIT_VARIABLES.AKIT_CONFIG_USER = filename
    return

def override_debug_breakpoints(breakpoints: List[str]):
    """
        This override function provides a mechanism overriding the AKIT_BREAKPOINTS
        variable and context configuration setting.

        :param breakpoints: A list of wellknown breakpoints that have been activated.
    """
    ctx.insert(ContextPaths.DEBUG_BREAKPOINTS, breakpoints)
    AKIT_VARIABLES.AKIT_BREAKPOINTS = breakpoints
    return

def override_debug_debugger(debugger: str):
    """
        This override function provides a mechanism overriding the AKIT_DEBUGGER
        variable and context configuration setting.

        :param debugger: The name of the debugger to setup.
    """
    ctx.insert(ContextPaths.DEBUG_DEBUGGER, debugger)
    AKIT_VARIABLES.AKIT_DEBUGGER = debugger
    return

def override_loglevel_console(level: str):
    """
        This override function provides a mechanism overriding the AKIT_LOG_LEVEL_CONSOLE
        variable and context configuration setting.

        :param level: The console logging level.
    """
    ctx.insert(ContextPaths.LOGGING_LEVEL_CONSOLE, level)
    AKIT_VARIABLES.AKIT_LOG_LEVEL_CONSOLE = level
    return

def override_loglevel_file(level: str):
    """
        This override function provides a mechanism overriding the AKIT_LOG_LEVEL_FILE
        variable and context configuration setting.

        :param level: The file log level.
    """
    ctx.insert(ContextPaths.LOGGING_LEVEL_LOGFILE, level)
    AKIT_VARIABLES.AKIT_LOG_LEVEL_FILE = level
    return

def override_job_initiator(job_initiator: str):
    """
        This override function provides a mechanism overriding the AKIT_JOB_INITIATOR
        variable and context configuration setting.

        :param job_initiator: The name of the initiator of the job.
    """

    ctx.insert(ContextPaths.JOB_INITIATOR, job_initiator)
    AKIT_VARIABLES.AKIT_JOB_INITIATOR = job_initiator
    
    return

def override_job_label(job_label: str):
    """
        This override function provides a mechanism overriding the AKIT_JOB_LABEL
        variable and context configuration setting.

        :param job_label: The name of the label if any the job was run under.
    """

    ctx.insert(ContextPaths.JOB_LABEL, job_label)
    AKIT_VARIABLES.AKIT_JOB_LABEL = job_label
    
    return

def override_job_name(job_name: str):
    """
        This override function provides a mechanism overriding the AKIT_JOB_NAME
        variable and context configuration setting.

        :param job_name: The name of the job.
    """

    ctx.insert(ContextPaths.JOB_NAME, job_name)
    AKIT_VARIABLES.AKIT_JOB_NAME = job_name
    
    return


def override_job_owner(job_owner: str):
    """
        This override function provides a mechanism overriding the AKIT_JOB_OWNER
        variable and context configuration setting.

        :param job_owner: The name of the owner of the job.
    """

    ctx.insert(ContextPaths.JOB_OWNER, job_owner)
    AKIT_VARIABLES.AKIT_JOB_OWNER = job_owner
    
    return

def override_output_directory(output_directory: str):
    """
        This override function provides a mechanism overriding the AKIT_STARTTIME
        variable and context configuration setting.

        :param output_directory: The base directory that all output artifacts will be written under.
    """
    ctx.insert(ContextPaths.OUTPUT_DIRECTORY, output_directory)
    AKIT_VARIABLES.AKIT_OUTPUT_DIRECTORY = output_directory
    return

def override_runid(runid: str):
    """
        This override function provides a mechanism overriding the AKIT_RUNID
        variable and context configuration setting.

        :param runid: A uuid string that represents the instance of this automation run.
    """
    ctx.insert(ContextPaths.RUNID, runid)
    AKIT_VARIABLES.AKIT_RUNID = runid
    return

def override_starttime(starttime: datetime):
    """
        This override function provides a mechanism overriding the AKIT_STARTTIME
        variable and context configuration setting.

        :param starttime: The date and time to set as the starttime of this run.
    """
    ctx.insert(ContextPaths.STARTTIME, starttime)
    AKIT_VARIABLES.AKIT_STARTTIME = starttime
    return

def override_testroot(testroot: str):
    """
        This override function provides a mechanism overriding the AKIT_TESTROOT
        variable and context configuration setting.

        :param testroot: The full path of the root of the tests folder.
    """
    AKIT_VARIABLES.AKIT_TESTROOT = testroot
    ctx.insert(ContextPaths.TESTROOT, testroot)
    return

def override_timetravel(allowed: bool):
    """
        This override function provides a mechanism overriding the AKIT_TIMETRAVEL
        variable and context configuration setting.
    """
    ctx.insert(ContextPaths.TIMETRAVEL, allowed)
    AKIT_VARIABLES.AKIT_TIMETRAVEL = allowed
    return

def override_timeportals(active_portals: List[str]):
    """
        This override function provides a mechanism overriding the AKIT_TIMEPORTALS
        variable and context configuration setting.
    """
    ctx.insert(ContextPaths.TIMEPORTALS, active_portals)
    AKIT_VARIABLES.AKIT_TIMEPORTALS = active_portals
    return




"""
.. module:: recorders
    :platform: Darwin, Linux, Unix, Windows
    :synopsis: Module which contains the :class:`TaskBase` object which is used as the base.

.. moduleauthor:: Myron Walker <myron.walker@gmail.com>
"""

__author__ = "Myron Walker"
__copyright__ = "Copyright 2020, Myron W Walker"
__credits__ = []
__version__ = "1.0.0"
__maintainer__ = "Myron Walker"
__email__ = "myron.walker@gmail.com"
__status__ = "Development" # Prototype, Development or Production
__license__ = "MIT"

from typing import Optional
from types import TracebackType

import collections
import datetime
import json
import os
import requests
import shutil

from datetime import datetime

from akit.exceptions import AKitNotOverloadedError

from akit.jsos import CHAR_RECORD_SEPERATOR
from akit.paths import (
    get_summary_html_template_source,
    get_summary_static_resource_dest_dir,
    get_summary_static_resource_src_dir
)
from akit.results import ResultCode, ResultNode, ResultType

from akit.testing.utilities import catalog_tree


class ResultRecorder:
    """
        The :class:`ResultRecorder` object is the base class object that establishes the API patterns
        for recorders of different formats to use when implementing a test result recorder.
    """
    def __init__(self, *, title: str, runid: str, start: datetime, summary_filename: str,
                 result_filename: str, apod: Optional[str] = None, branch: Optional[str] = None,
                 build: Optional[str] = None, flavor: Optional[str] = None, build_url: Optional[str] = None,
                 job_initiator: Optional[str] = None, job_label: Optional[str] = None, job_name: Optional[str] = None,
                 job_owner: Optional[str] = None, job_type: Optional[str] = None):
        """
            Initializes an instance of a ResultRecorder with the information about a test run.

            :param title: A title to associated with the summary for the test results.
            :param runid: The uuid string that identifies a set of test results.
            :param start: The date and time of the start of the test run.
            :param summary_filename: The full path to the summary file where the test run summary should be written to.
            :param result_filename: The full path to the results file where the test run results should be written to.
            :param apod: Optional name of an automation pod that the testrun is running on.
            :param branch: Optional name of a code 'branch' to associate with the test results.
            :param build: Optional name of a product 'build' to associate with the test results.
            :param flavor: Optional label that indicates the flavor of build the test run is running against.
            :param build_url: Optional build url
            :param job_initiator: Optional name of the initiator of the job.
            :param job_label: Optional label associated with the job.
            :param job_name: Optional name of the job.
            :param job_owner: Optional owner of the job.
            :param job_type: Optional job type.
        """

        self._title = title
        self._runid = runid
        self._start = start
        self._summary_filename = summary_filename
        self._result_filename = result_filename
        self._apod = apod
        self._branch = branch
        self._build = build
        self._flavor = flavor
        self._build_url = build_url
        self._job_initiator = job_initiator
        self._job_label = job_label
        self._job_name = job_name
        self._job_owner = job_owner
        self._job_type = job_type

        self._output_dir = os.path.dirname(summary_filename)

        self._rout = None

        self._error_count = 0
        self._failure_count = 0
        self._pass_count = 0
        self._skip_count = 0
        self._unknown_count = 0

        self._total_count = 0

        self._finalized = False

        build_info = collections.OrderedDict((
            ("branch", branch),
            ("build", build),
            ("flavor", flavor)
        ))

        job_info = collections.OrderedDict((
            ("initiator", job_initiator),
            ("label", job_label),
            ("name", job_name),
            ("owner", job_owner),
            ("type", job_type)
        ))

        self._summary = collections.OrderedDict((
            ("title", self._title),
            ("runid", self._runid),
            ("build", build_info),
            ("job", job_info),
            ("start", self._start),
            ("stop", None),
            ("result", "RUNNING"),
            ("apod", self._apod),
            ("detail", None)
        ))
        return

    def __enter__(self):
        """
            Starts up the recording process of test results.
        """
        self.update_summary()
        self._rout = open(self._result_filename, 'w')
        return self

    def __exit__(self, ex_type: type, ex_inst: Exception, ex_tb: TracebackType) -> bool:
        """
            Starts up the recording process of test results.

            :param ex_type: The type associated with the exception being raised.
            :param ex_inst: The exception instance of the exception being raised.
            :param ex_tb: The traceback associated with the exception being raised.

            :returns: Returns true if an exception was handled and should be suppressed.
        """
        if not self._finalized:
            self.finalize()
        return

    @property
    def summary(self):
        """
            Get the result summary.
        """
        return self._summary

    def finalize(self):
        """
            Finalizes the test results counters and status of the test run.
        """
        self._finalized = True

        self._stop = datetime.now()
        self._summary["stop"] = str(self._stop)

        self._summary["detail"] = {
            "errors": self._error_count,
            "failed": self._failure_count,
            "skipped": self._skip_count,
            "passed": self._pass_count,
            "total": self._total_count
        }

        if self._error_count > 0 or self._failure_count > 0:
            self._summary["result"] = "FAILED"
        else:
            self._summary["result"] = "PASSED"

        self._rout.close()

        self.update_render_environment()

        summary_html_source = get_summary_html_template_source()
        summary_render_html_base = self.get_summary_render_file_basename()
        summary_render_html_dest = os.path.join(self._output_dir, summary_render_html_base)
        catalog_tree(self._output_dir, ignore_dirs=["__pycache__"])
        shutil.copy2(summary_html_source, summary_render_html_dest)

        self.update_summary()

        return
    
    def format_lines(self):
        lines = [
            " ============== Test Summary ============== ",
        ]

        if self._title:
            lines.append("   Title: {}".format(self._title))
        if self._branch:
            lines.append("  Branch: {}".format(self._branch))
        if self._build:
            lines.append("   Build: {}".format(self._build))
        if self._flavor:
            lines.append("  Flavor: {}".format(self._flavor))
        if self._job_owner:
            lines.append("   Owner: {}".format(self._job_owner))

        lines.extend([
            "   RunId: {}".format(self._runid),
            "   Start: {}".format(self._start),
            "    Stop: {}".format(self._stop),
            " ----------------- Detail ----------------- ",
            "       Errors: {}".format(self._error_count),
            "       Failed: {}".format(self._failure_count),
            "      Skipped: {}".format(self._skip_count),
            "       Passed: {}".format(self._pass_count),
            "        Total: {}".format(self._total_count),
            " ========================================== ",
            "   {}".format(self._summary["result"]),
            " ========================================== ",
            "",
            "OUTPUT PATH: {}".format(self._output_dir)
        ])

        if "AKIT_SUMMARY_URL" in os.environ:
            summary_url = os.environ["AKIT_SUMMARY_URL"]
            lines.append("SUMMARY URL: {}".format(summary_url))

        return lines

    def get_summary_render_file_basename(self):
        return "testsummary.html"

    def record(self, result: ResultNode):
        """
            Records an entry for the result object that is passed.

            :param result: A result object to be recorded.
        """
        if result.result_type == ResultType.TEST:
            self._total_count += 1

            result_code = result.result_code
            if result_code == ResultCode.PASSED:
                self._pass_count += 1
            elif result_code == ResultCode.ERRORED:
                self._error_count += 1
            elif result_code == ResultCode.FAILED:
                self._failure_count += 1
            elif result_code == ResultCode.SKIPPED:
                self._skip_count += 1
            else:
                self._unknown_count += 1

        json_str = result.to_json()

        self._rout.write(CHAR_RECORD_SEPERATOR)
        self._rout.write(json_str)
        return

    def update_render_environment(self):
        """
            Called in order to publish resources required by the HTML presentation to
            the results directory tree.
            
            ..note: For Jenkins deployments and runs, you might want to override this method and
                    not publish for each run.
        """

        static_resource_dest_dir = get_summary_static_resource_dest_dir()
        static_resource_src_dir = get_summary_static_resource_src_dir()

        for nxt_root, _, nxt_files in os.walk(static_resource_src_dir):
            for nf in nxt_files:
                res_src_full = os.path.join(nxt_root, nf)
                res_src_leaf = res_src_full[len(static_resource_src_dir):].lstrip(os.sep)
                res_dest_full = os.path.join(static_resource_dest_dir, res_src_leaf)
                if not os.path.exists(res_dest_full):
                    dest_dir = os.path.dirname(res_dest_full)
                    if not os.path.exists(dest_dir):
                        os.makedirs(dest_dir)
                    shutil.copy2(res_src_full, res_dest_full)

        return

    def update_summary(self): # pylint: disable=no-self-use
        """
            Writes out an update to the test run summary file.
        """
        raise AKitNotOverloadedError("The 'update_summary' method must be overridden by derived 'ResultRecorder' objects.") from None


class JsonResultRecorder(ResultRecorder):
    """
        The :class:`JsonResultRecorder` object records test results in JSON format.
    """
    def __init__(self, *, title: str, runid: str, start: datetime, summary_filename: str,
                 result_filename: str, apod: Optional[str] = None, branch: Optional[str] = None,
                 build: Optional[str] = None, flavor: Optional[str] = None, build_url: Optional[str] = None,
                 job_initiator: Optional[str] = None, job_label: Optional[str] = None, job_name: Optional[str] = None,
                 job_owner: Optional[str] = None, job_type: Optional[str] = None):
        """
            Initializes the :class:`JsonResultRecorder` object for recording test results for
            a test run.

            :param title: A title to associated with the summary for the test results.
            :param runid: The uuid string that identifies a set of test results.
            :param start: The date and time of the start of the test run.
            :param summary_filename: The full path to the summary file where the test run summary should be written to.
            :param result_filename: The full path to the results file where the test run results should be written to.
            :param apod: Optional name of an automation pod that the testrun is running on.
            :param branch: Optional name of a code 'branch' to associate with the test results.
            :param build: Optional name of a product 'build' to associate with the test results.
            :param flavor: Optional label that indicates the flavor of build the test run is running against.
            :param build_url: Optional build url
            :param job_initiator: Optional name of the initiator of the job.
            :param job_label: Optional label associated with the job.
            :param job_name: Optional name of the job.
            :param job_owner: Optional owner of the job.
            :param job_type: Optional job type.
        """
        super(JsonResultRecorder, self).__init__(title=title, runid=runid, start=start, summary_filename=summary_filename,
            result_filename= result_filename, apod=apod, branch=branch, build=build, flavor=flavor, build_url=build_url,
            job_initiator=job_initiator, job_label=job_label, job_name=job_name, job_owner=job_owner,
            job_type=job_type)
        return

    def update_summary(self):
        """
            Writes out an update to the test run summary file.
        """

        with open(self._summary_filename, 'w') as sout:
            json.dump(self._summary, sout, indent=4)

        return



"""

    NOTE: This is a code generated file.  This file should not be edited directly.
"""



from akit.aspects import AspectsUPnP, DEFAULT_UPNP_ASPECTS

from akit.extensible import LoadableExtension
from akit.interop.upnp.services.upnpserviceproxy import UpnpServiceProxy

class DeviceProtection1ServiceProxy(UpnpServiceProxy, LoadableExtension):
    """
        This is a code generated proxy class to the 'urn:schemas-upnp-org:service:DeviceProtection:1' service.
    """

    SERVICE_MANUFACTURER = 'UPnP'
    SERVICE_TYPE = 'DeviceProtection1'

    SERVICE_DEFAULT_VARIABLES = {
        "SetupReady": { "data_type": "boolean", "default": None, "allowed_list": None},
        "SupportedProtocols": { "data_type": "string", "default": None, "allowed_list": None},
    }

    SERVICE_EVENT_VARIABLES = {}

    def action_AddIdentityList(self, IdentityList, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the AddIdentityList action.

            :returns: "IdentityListResult"
        """
        arguments = {
            "IdentityList": IdentityList,
        }

        out_params = self.call_action("AddIdentityList", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("IdentityListResult",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_AddRolesForIdentity(self, Identity, RoleList, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the AddRolesForIdentity action.
        """
        arguments = {
            "Identity": Identity,
            "RoleList": RoleList,
        }

        self.call_action("AddRolesForIdentity", arguments=arguments, aspects=aspects)

        return

    def action_GetACLData(self, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetACLData action.

            :returns: "ACL"
        """
        arguments = { }

        out_params = self.call_action("GetACLData", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("ACL",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetAssignedRoles(self, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetAssignedRoles action.

            :returns: "RoleList"
        """
        arguments = { }

        out_params = self.call_action("GetAssignedRoles", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("RoleList",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetRolesForAction(self, DeviceUDN, ServiceId, ActionName, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetRolesForAction action.

            :returns: "RoleList", "RestrictedRoleList"
        """
        arguments = {
            "DeviceUDN": DeviceUDN,
            "ServiceId": ServiceId,
            "ActionName": ActionName,
        }

        out_params = self.call_action("GetRolesForAction", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("RoleList", "RestrictedRoleList",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetSupportedProtocols(self, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetSupportedProtocols action.

            :returns: "ProtocolList"
        """
        arguments = { }

        out_params = self.call_action("GetSupportedProtocols", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("ProtocolList",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetUserLoginChallenge(self, ProtocolType, Name, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetUserLoginChallenge action.

            :returns: "Salt", "Challenge"
        """
        arguments = {
            "ProtocolType": ProtocolType,
            "Name": Name,
        }

        out_params = self.call_action("GetUserLoginChallenge", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("Salt", "Challenge",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_RemoveIdentity(self, Identity, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the RemoveIdentity action.
        """
        arguments = {
            "Identity": Identity,
        }

        self.call_action("RemoveIdentity", arguments=arguments, aspects=aspects)

        return

    def action_RemoveRolesForIdentity(self, Identity, RoleList, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the RemoveRolesForIdentity action.
        """
        arguments = {
            "Identity": Identity,
            "RoleList": RoleList,
        }

        self.call_action("RemoveRolesForIdentity", arguments=arguments, aspects=aspects)

        return

    def action_SendSetupMessage(self, ProtocolType, InMessage, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the SendSetupMessage action.

            :returns: "OutMessage"
        """
        arguments = {
            "ProtocolType": ProtocolType,
            "InMessage": InMessage,
        }

        out_params = self.call_action("SendSetupMessage", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("OutMessage",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_SetUserLoginPassword(self, ProtocolType, Name, Stored, Salt, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the SetUserLoginPassword action.
        """
        arguments = {
            "ProtocolType": ProtocolType,
            "Name": Name,
            "Stored": Stored,
            "Salt": Salt,
        }

        self.call_action("SetUserLoginPassword", arguments=arguments, aspects=aspects)

        return

    def action_UserLogin(self, ProtocolType, Challenge, Authenticator, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the UserLogin action.
        """
        arguments = {
            "ProtocolType": ProtocolType,
            "Challenge": Challenge,
            "Authenticator": Authenticator,
        }

        self.call_action("UserLogin", arguments=arguments, aspects=aspects)

        return

    def action_UserLogout(self, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the UserLogout action.
        """
        arguments = { }

        self.call_action("UserLogout", arguments=arguments, aspects=aspects)

        return



__author__ = "Myron Walker"
__copyright__ = "Copyright 2020, Myron W Walker"
__credits__ = []
__version__ = "1.0.0"
__maintainer__ = "Myron Walker"
__email__ = "myron.walker@gmail.com"
__status__ = "Development" # Prototype, Development or Production
__license__ = "MIT"

from typing import Callable, Dict, Optional

import inspect

from akit.exceptions import AKitSemanticError
from akit.coupling.integrationcoupling import IntegrationCoupling

from akit.testing.testplus.registration.resourceregistry import resource_registry

from akit.testing.testplus.registration.integrationsource import IntegrationSource
from akit.testing.testplus.resourcelifespan import ResourceLifespan
from akit.testing.testplus.registration.parameterorigin import ParameterOrigin

def param(source, *, identifier: Optional[None], constraints: Optional[Dict]=None):
    def decorator(subscriber: Callable) -> Callable:
        nonlocal source
        nonlocal identifier
        nonlocal constraints

        if identifier is None:
            identifier = source.__name__

        if identifier == 'constraints':
            errmsg = "Invalid identifier.  The word 'constraints' is reseved for delivering dynamic constraints."
            raise AKitSemanticError(errmsg) from None

        life_span = ResourceLifespan.Test

        source_info = resource_registry.lookup_resource_source(source)

        if constraints is not None and 'constraints' not in source_info.source_signature.parameters:
            raise AKitSemanticError("Attempting to pass constraints to a parameter origin with no 'constraints' parameter.") from None

        assigned_scope = "{}#{}".format(subscriber.__module__, subscriber.__name__)

        param_origin = ParameterOrigin(assigned_scope, identifier, source_info, life_span, constraints)
        resource_registry.register_parameter_origin(identifier, param_origin)

        return subscriber
    return decorator

def originate_parameter(source_func, *, identifier: Optional[None], life_span: ResourceLifespan=None, assigned_scope: Optional[str]=None, constraints: Optional[Dict]=None):

    if source_func is None:
        errmsg = "The 'source_func' parameter cannot be 'None'."
        raise AKitSemanticError(errmsg) from None

    if life_span == ResourceLifespan.Test:
        errmsg = "The 'life_span' parameter cannot be 'ResourceLifespan.Test'."
        raise AKitSemanticError(errmsg) from None

    if identifier is None:
        identifier = source_func.__name__

    if identifier == 'constraints':
        errmsg = "Invalid identifier.  The word 'constraints' is reseved for delivering dynamic constraints."
        raise AKitSemanticError(errmsg) from None

    source_info = resource_registry.lookup_resource_source(source_func)
    if assigned_scope is not None:
        if isinstance(source_info, IntegrationSource):
            errmsg = "The 'assigned_scope' parameter should not be specified unless the source of the resource is of type 'scope' or 'resource'."
            raise AKitSemanticError(errmsg) from None

    if constraints is not None and 'constraints' not in source_info.source_signature.parameters:
            raise AKitSemanticError("Attempting to pass constraints to a parameter origin with no 'constraints' parameter.") from None

    caller_frame = inspect.stack()[1]
    calling_module = inspect.getmodule(caller_frame[0])
    
    if life_span is None:
        res_type = source_info.resource_type
        if issubclass(res_type, IntegrationCoupling):
            life_span = ResourceLifespan.Session
        else:
            life_span = ResourceLifespan.Package

    if life_span == ResourceLifespan.Package:
        if assigned_scope is None:
            assigned_scope = calling_module.__name__
    elif life_span == ResourceLifespan.Session:
        assigned_scope = "<session>"

    param_origin = ParameterOrigin(assigned_scope, identifier, source_info, life_span, constraints)
    resource_registry.register_parameter_origin(identifier, param_origin)

    return





from typing import Callable, Type

import inspect

class SourceBase:

    def __init__(self, source_func: Callable, query_func: Callable, resource_type: Type, constaints: dict):
        self._source_func = source_func
        self._query_func = query_func
        self._resource_type = resource_type
        self._constraints = constaints
        self._subscriptions = None
        return

    @property
    def constraints(self):
        return self._constraints

    @property
    def module_name(self) -> str:
        return self._source_func.__module__

    @property
    def query_function(self) -> Callable:
        return self._query_func

    @property
    def resource_type(self) -> Type:
        return self._resource_type

    @property
    def source_function(self) -> Callable:
        return self._source_func

    @property
    def source_signature(self) -> inspect.Signature:
        return inspect.signature(self._source_func)

    @property
    def source_id(self) -> str:
        source = self._source_func
        idstr = "{}#{}".format(source.__module__, source.__name__)
        return idstr

    @property
    def subscriptions(self):
        return self._subscriptions

    @subscriptions.setter
    def subscriptions(self, val):
        self._subscriptions = val
        return


__author__ = "Myron Walker"
__copyright__ = "Copyright 2020, Myron W Walker"
__credits__ = []
__version__ = "1.0.0"
__maintainer__ = "Myron Walker"
__email__ = "myron.walker@gmail.com"
__status__ = "Development" # Prototype, Development or Production
__license__ = "MIT"

import os

from akit.environment.variables import LOG_LEVEL_NAMES

import click

import yaml

HELP_WORK = "The file containing the workpacket detail for performing a workflow orchestration."
HELP_OUTPUT = "The output directory where results and artifacts are collected."
HELP_START = r"A time stamp to associate with the start of the run. Example: 2020-10-17T15:30:11.989120  Bash: date +%Y-%m-%dT%H:%M:%S.%N"
HELP_CONSOLE_LOG_LEVEL = "The logging level for console output."
HELP_FILE_LOG_LEVEL = "The logging level for logfile output."

@click.group("workflow", help="Contains commands that are used to work with or run workflow scripts.")
def group_akit_workflow():
    return

@click.command("run")
@click.argument("workflow", required=True)
@click.option("--output", "-o", required=False, help=HELP_OUTPUT)
@click.option("--start", default=None, required=False, help=HELP_START)
@click.option("--console-level", default=None, required=False, type=click.Choice(LOG_LEVEL_NAMES, case_sensitive=False), help=HELP_CONSOLE_LOG_LEVEL)
@click.option("--logfile-level", default=None, required=False, type=click.Choice(LOG_LEVEL_NAMES, case_sensitive=False), help=HELP_FILE_LOG_LEVEL)
def command_akit_workflow_run(workflow, output=None, start=None, console_level=None, logfile_level=None):

    # pylint: disable=unused-import,import-outside-toplevel

    # We do the imports of the automation framework code inside the action functions because
    # we don't want to startup loggin and the processing of inputs and environment variables
    # until we have entered an action method.  Thats way we know how to setup the environment.

    # IMPORTANT: We need to load the context first because it will trigger the loading
    # of the default user configuration
    from akit.environment.context import Context
    from akit.environment.variables import extend_path

    ctx = Context()
    env = ctx.lookup("/environment")

    workflow_file = os.path.abspath(os.path.expanduser(os.path.expandvars(workflow)))
    if not os.path.exists(workflow_file):
        error_msg = "The specified workflow file does not exist. file=%s" % workflow_file
        raise click.BadParameter(error_msg)

    workflow_info = None
    with open(workflow_file, 'r') as wpf:
        wpfcontent = wpf.read()
        workflow_info = yaml.safe_load(wpfcontent)

    if workflow_info is not None:

        from akit.workflow.entrypoints import run_workflow_entrypoint

        # Run the work packet
        run_workflow_entrypoint(workflow_file, workflow_info)

    else:
        error_msg = "Failure loading the work packet info from. file=%s" % workflow_file
        raise click.BadParameter(error_msg)

    return

group_akit_workflow.add_command(command_akit_workflow_run)



__author__ = "Myron Walker"
__copyright__ = "Copyright 2020, Myron W Walker"
__credits__ = []
__version__ = "1.0.0"
__maintainer__ = "Myron Walker"
__email__ = "myron.walker@gmail.com"
__status__ = "Development" # Prototype, Development or Production
__license__ = "MIT"

from typing import Optional
from types import ModuleType

def parse_test_include_expression(expression: str, testmodule: Optional[ModuleType], method_prefix: str = "test"):
    """
        Parses the test include expression in connection with the testmodule and method_prefix information provided
        and returns the components (package, module, testclass, testname) which can be used to perform a
        test tree search for included tests based on the expression.

        :param expression: A test include expression in the form of (package).(module)@(testclass)#(testcase).  The module,
                           testclass, and testcase are optional.  If these items are excluded every descendant test found
                           under the (package) will be included.
        :param testmodule: A test module that contains the tests to be run.  This is passed when a individual test file
                           debugging workflow is being used.
        :param method_prefix: The string prefix that identifies test methods on a :class:`TestContainer` derived class.
    """

    expr_package = None
    expr_module = None
    expr_testclass = None
    expr_testname = None

    # If test_module was passed then we are running a test module as a script or debugging a test module
    # so we handle the special case where we only collect the test references from the test module that
    # was set.
    if testmodule is not None:
        if expression.find(".") > -1 or expression.find("@") > -1 or expression.find(":") > -1:
            raise ValueError("parse_test_include_expression: The include expression for test module runs should only have a " \
                             "test class and test method.")

        testmodule_name = testmodule.__name__

        expr_package = "*"
        expr_module = "*"

        # If the full path to the module was set correctly by the generic entry point, then replace the package and module
        # expressions with exact expressions so we don't load coad that does not need to be loaded when we are scanning
        # for tests.
        if testmodule_name != "__main__":
            tm_name_parts = testmodule_name.split('.')
            if len(tm_name_parts) > 1:
                expr_package = '.'.join(tm_name_parts[:-1])
                expr_module = tm_name_parts[-1]

        if expression.find("#") > -1:
            expression, expr_testname = expression.split("#")
            if not expr_testname.startswith(method_prefix):
                raise ValueError("parse_test_include_expression: The testname component of the expression must start with the " \
                                "method_prefix=%r. expression%r" % (method_prefix, expression))

        if expression == "*":
            expr_testclass = None

    # If self._test_module was not set then we are performing a commmandline run where a test job or includes, excludes
    # collection was passed we need to use one of those to determine what to run.
    else:
        # Start from the end of the expression and work backwards to determine what components in the expression were passed

        # First look for the # to see if we have a test name specified.
        if expression.find("#") > -1:
            expression, expr_testname = expression.split("#")
            if not expr_testname.startswith(method_prefix):
                raise ValueError("parse_test_include_expression: The testname component of the expression must " \
                                "start with the method_prefix=%r. expression%r" % (method_prefix, expression))

        if expression.find("@") > -1:
            expression, expr_testclass = expression.split("@")

        if expr_testclass == "":
            expr_testclass = None

        comb_expr_comp = expression.split(".")
        if len(comb_expr_comp) > 1:
            expr_package = ".".join(comb_expr_comp[:-1])
            expr_module = comb_expr_comp[-1]
        else:
            expr_package = None
            expr_module = expression

    return expr_package, expr_module, expr_testclass, expr_testname



"""

    NOTE: This is a code generated file.  This file should not be edited directly.
"""



from akit.aspects import AspectsUPnP, DEFAULT_UPNP_ASPECTS

from akit.extensible import LoadableExtension
from akit.interop.upnp.services.upnpserviceproxy import UpnpServiceProxy

class RADASync2ServiceProxy(UpnpServiceProxy, LoadableExtension):
    """
        This is a code generated proxy class to the 'urn:schemas-upnp-org:service:RADASync:2' service.
    """

    SERVICE_MANUFACTURER = 'UPnP'
    SERVICE_TYPE = 'RADASync2'

    SERVICE_DEFAULT_VARIABLES = {
        "SystemInfo": { "data_type": "string", "default": None, "allowed_list": None},
    }

    SERVICE_EVENT_VARIABLES = {}

    def action_AddRemoteDevices(self, DeviceList, ID, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the AddRemoteDevices action.
        """
        arguments = {
            "DeviceList": DeviceList,
            "ID": ID,
        }

        self.call_action("AddRemoteDevices", arguments=arguments, aspects=aspects)

        return

    def action_GetLocalNetworkAddressInfo(self, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetLocalNetworkAddressInfo action.

            :returns: "LocalNetworkAddress"
        """
        arguments = { }

        out_params = self.call_action("GetLocalNetworkAddressInfo", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("LocalNetworkAddress",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_HeartbeatUpdate(self, ID, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the HeartbeatUpdate action.
        """
        arguments = {
            "ID": ID,
        }

        self.call_action("HeartbeatUpdate", arguments=arguments, aspects=aspects)

        return

    def action_RemoveRemoteDevices(self, DeviceList, ID, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the RemoveRemoteDevices action.
        """
        arguments = {
            "DeviceList": DeviceList,
            "ID": ID,
        }

        self.call_action("RemoveRemoteDevices", arguments=arguments, aspects=aspects)

        return

    def action_SetDDDLocation(self, DDDLocation, ID, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the SetDDDLocation action.
        """
        arguments = {
            "DDDLocation": DDDLocation,
            "ID": ID,
        }

        self.call_action("SetDDDLocation", arguments=arguments, aspects=aspects)

        return



"""

    NOTE: This is a code generated file.  This file should not be edited directly.
"""



from akit.aspects import AspectsUPnP, DEFAULT_UPNP_ASPECTS

from akit.extensible import LoadableExtension
from akit.interop.upnp.services.upnpserviceproxy import UpnpServiceProxy

class Presence1ServiceProxy(UpnpServiceProxy, LoadableExtension):
    """
        This is a code generated proxy class to the 'urn:schemas-upnp-org:service:Presence:1' service.
    """

    SERVICE_MANUFACTURER = 'UPnP'
    SERVICE_TYPE = 'Presence1'

    SERVICE_DEFAULT_VARIABLES = {}

    SERVICE_EVENT_VARIABLES = {
        "PresenceOfContactsUpdate": { "data_type": "string", "default": None, "allowed_list": None},
        "UserPresenceInfo": { "data_type": "string", "default": None, "allowed_list": None},
        "Watcher": { "data_type": "string", "default": None, "allowed_list": None},
    }

    def action_AuthorizePresenceProactive(self, UserPresenceInfo, Expire, WatcherList, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the AuthorizePresenceProactive action.
        """
        arguments = {
            "UserPresenceInfo": UserPresenceInfo,
            "Expire": Expire,
            "WatcherList": WatcherList,
        }

        self.call_action("AuthorizePresenceProactive", arguments=arguments, aspects=aspects)

        return

    def action_AuthorizePresenceReactive(self, Contact, Expire, UserPresenceInfo, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the AuthorizePresenceReactive action.
        """
        arguments = {
            "Contact": Contact,
            "Expire": Expire,
            "UserPresenceInfo": UserPresenceInfo,
        }

        self.call_action("AuthorizePresenceReactive", arguments=arguments, aspects=aspects)

        return

    def action_GetContactPresence(self, TargetContact, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetContactPresence action.

            :returns: "ContactPresence"
        """
        arguments = {
            "TargetContact": TargetContact,
        }

        out_params = self.call_action("GetContactPresence", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("ContactPresence",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetPresence(self, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetPresence action.

            :returns: "UserPresence"
        """
        arguments = { }

        out_params = self.call_action("GetPresence", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("UserPresence",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetPresenceOfContactsUpdate(self, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetPresenceOfContactsUpdate action.

            :returns: "ContactPresenceUpdate"
        """
        arguments = { }

        out_params = self.call_action("GetPresenceOfContactsUpdate", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("ContactPresenceUpdate",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_RegisterForContactPresence(self, Contact, Expire, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the RegisterForContactPresence action.

            :returns: "RegistrationResult"
        """
        arguments = {
            "Contact": Contact,
            "Expire": Expire,
        }

        out_params = self.call_action("RegisterForContactPresence", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("RegistrationResult",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_UpdatePresence(self, UpdatedUserPresence, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the UpdatePresence action.
        """
        arguments = {
            "UpdatedUserPresence": UpdatedUserPresence,
        }

        self.call_action("UpdatePresence", arguments=arguments, aspects=aspects)

        return



__author__ = "Myron Walker"
__copyright__ = "Copyright 2020, Myron W Walker"
__credits__ = []
__version__ = "1.0.0"
__maintainer__ = "Myron Walker"
__email__ = "myron.walker@gmail.com"
__status__ = "Development" # Prototype, Development or Production
__license__ = "MIT"

TIMEDELTA_RENEWAL_WINDOW = 120

from akit.aspects import AspectsUPnP

UPNP_CALL_COMPLETION_TIMEOUT = 30
UPNP_CALL_COMPLETION_INTERVAL = 5

UPNP_CALL_INACTIVITY_TIMEOUT = 30
UPNP_CALL_INACTIVITY_INTERVAL = 1

DEFAULT_UPNP_CALL_ASPECTS = AspectsUPnP(
    completion_timeout=UPNP_CALL_COMPLETION_TIMEOUT,
    completion_interval=UPNP_CALL_COMPLETION_INTERVAL,
    inactivity_timeout=UPNP_CALL_INACTIVITY_TIMEOUT,
    inactivity_interval=UPNP_CALL_INACTIVITY_INTERVAL)


"""
.. module:: integration
    :platform: Darwin, Linux, Unix, Windows
    :synopsis: Module containing the :class:`IntegrationCoupling` class and associated reflection methods.
        The :class:`IntegrationCoupling` derived classes can be used to integraton automation resources and roles
        into the test environment.

.. moduleauthor:: Myron Walker <myron.walker@gmail.com>
"""

__author__ = "Myron Walker"
__copyright__ = "Copyright 2020, Myron W Walker"
__credits__ = []
__version__ = "1.0.0"
__maintainer__ = "Myron Walker"
__email__ = "myron.walker@gmail.com"
__status__ = "Development" # Prototype, Development or Production
__license__ = "MIT"

from typing import Dict, List, Tuple, TYPE_CHECKING

import inspect

from akit.exceptions import AKitNotOverloadedError

from akit.coupling.basecoupling import BaseCoupling

from akit.xlogging.foundations import getAutomatonKitLogger

# Types imported only for type checking purposes
if TYPE_CHECKING:
    from akit.interop.landscaping.landscape import Landscape

class IntegrationCoupling(BaseCoupling):
    """
        The :class:`IntegrationCoupling` object serves as the base object for the declaration of an
        automation integration requirement.  The :class:`akit.testing.unittest.testsequencer.Sequencer`
        queries the class hierarchies of the tests that are included in an automation run.
    """

    landscape = None

    logger = getAutomatonKitLogger()
    pathbase = None

    CONSTRAINTS = None

    def __init__(self, *args, role=None, **kwargs): # pylint: disable=unused-argument
        """
            The default contructor for an :class:`IntegrationCoupling`.
        """
        super(IntegrationCoupling, self).__init__()

        self._role = role
        self._connectivity_establish = False
        self._presence_establish = False

        if self.pathbase is None:
            err_msg = "The 'pathbase' class member variable must be set to a unique name for each integration class type."
            raise ValueError(err_msg)

        self.context.insert(self.pathbase, self)
        return

    @property
    def connectivity_establish(self):
        return self._connectivity_establish

    @property
    def mode(self):
        """
            Returns the current mode any associated resource is operating in.
        """
        return self._mode

    @mode.setter
    def mode(self, value):
        """
            Sets the current mode any associated resource is operating in.
        """
        old_value = self._mode
        self._mode = value
        self.on_mode_changed(old_value, value)
        return

    @property
    def presence_establish(self):
        return self._presence_establish

    @property
    def role(self):
        """
            Returns the current automation role assigned to the integration coupling.
        """
        return self._role

    def on_mode_changed(self, prev_mode, new_mode): # pylint: disable=no-self-use,unused-argument
        """
            Implemented by derived classes in order to perform the changeover of modes.
        """
        return

    @classmethod
    def declare_precedence(cls) -> int:
        """
            This API is called so that the IntegrationCoupling can declare an ordinal precedence that should be
            utilized for bringing up its integration state.
        """
        cls.logger = getAutomatonKitLogger()
        return

    @classmethod
    def attach_to_environment(cls, constraints: Dict={}):
        """
            This API is called so that the IntegrationCoupling can process configuration information.  The :class:`IntegrationCoupling`
            will verify that it has a valid environment and configuration to run in.

            :raises :class:`akit.exceptions.AKitMissingConfigError`, :class:`akit.exceptions.AKitInvalidConfigError`:
        """
        raise AKitNotOverloadedError("The 'attach_to_environment' method must be overloaded by derived integration coupling types.") from None
    
    @classmethod
    def attach_to_framework(cls, landscape: "Landscape"):
        """
            This API is called so that the IntegrationCoupling can attach to the test framework and participate with
            registration processes.  This allows the framework to ignore the bring-up of couplings that are not being
            included by a test.
        """
        cls.landscape = landscape
        return

    @classmethod
    def collect_resources(cls):
        """
            This API is called so the `IntegrationCoupling` can connect with a resource management
            system and gain access to the resources required for the automation run.

            :raises :class:`akit.exceptions.AKitResourceError`:
        """
        raise AKitNotOverloadedError("The 'collect_resources' method must be overloaded by derived integration coupling types.") from None

    @classmethod
    def diagnostic(cls, label: str, level: int, diag_folder: str): # pylint: disable=unused-argument
        """
            The API is called by the :class:`akit.sequencer.Sequencer` object when the automation sequencer is
            building out a diagnostic package at a diagnostic point in the automation sequence.  Example diagnostic
            points are:

            * pre-run
            * post-run

            Each diagnostic package has its own storage location so derived :class:`akit.scope.ScopeCoupling` objects
            can simply write to their specified output folder.

            :param label: The label associated with this diagnostic.
            :param level: The maximum diagnostic level to run dianostics for.
            :param diag_folder: The output folder path where the diagnostic information should be written.
        """
        return

    @classmethod
    def establish_connectivity(cls, allow_missing_devices: bool=False) -> Tuple[List[str], Dict]:
        """
            This API is called so the `IntegrationCoupling` can establish connectivity with any compute or storage
            resources.

            :returns: A tuple with a list of error messages for failed connections and dict of connectivity
                      reports for devices devices based on the coordinator.
        """
        raise AKitNotOverloadedError("The 'diagnostic' method must be overloaded by derived integration coupling types.") from None

    @classmethod
    def establish_presence(cls) -> Tuple[List[str], Dict]:
        """
            This API is called so the `IntegrationCoupling` can establish presence with any compute or storage
            resources.

            :returns: A tuple with a list of error messages for failed connections and dict of connectivity
                      reports for devices devices based on the coordinator.
        """
        raise AKitNotOverloadedError("The 'diagnostic' method must be overloaded by derived integration coupling types.") from None


def is_integration_coupling(cls):
    """
        Helper function that is used to determine if a type is an :class:`IntegrationCoupling` subclass, but not
        the :class:`IntegrationCoupling` type itself.
    """
    is_integmi = False
    if inspect.isclass(cls) and cls is not IntegrationCoupling and issubclass(cls, IntegrationCoupling):
        is_integmi = True
    return is_integmi

def sort_integration_couplings_by_precedence(self, coupling_list: List[IntegrationCoupling]) -> List[IntegrationCoupling]:
    """
        Takes a list of :class:`IntegrationCoupling` classes and creates an ordered list based on the ordinal
        precedence declared by the :class:`IntegrationCoupling`.
    """
    precedence_table = {}

    for coupling in coupling_list:
        precedence = coupling.declare_precedence()
        precedence_level_list = None
        if precedence in precedence_table:
            precedence_level_list = precedence_table[precedence]
        else:
            precedence_level_list = []
            precedence_table[precedence] = precedence_level_list
        precedence_level_list.append(coupling)

    precedence_keys_sorted = precedence_table.keys()
    precedence_keys_sorted.sort()

    ordered_couplings = []
    for precedence in precedence_keys_sorted:
        ordered_couplings.extend(precedence_table[precedence])

    return ordered_couplings


"""
.. module:: testref
    :platform: Darwin, Linux, Unix, Windows
    :synopsis: Module containing the base :class:`TestRef` type used to reference to
        tests that will be included into an a test execution graph.

.. moduleauthor:: Myron Walker <myron.walker@gmail.com>
"""

__author__ = "Myron Walker"
__copyright__ = "Copyright 2020, Myron W Walker"
__credits__ = []
__version__ = "1.0.0"
__maintainer__ = "Myron Walker"
__email__ = "myron.walker@gmail.com"
__status__ = "Development" # Prototype, Development or Production
__license__ = "MIT"

from typing import Any, Dict, List, OrderedDict, Optional, Sequence

from types import FunctionType

import collections
import inspect
import sys

from akit.testing.testplus.markers import MetaFilter

class TestRef:
    """
        The :class:`TestRef` objects are used to refer to a reference to a test.  We use :class:`TestRef` instances
        to reference the tests that are going to be run so we can control the lifespan of test case instances
        and control our memory consumption during test runs with large collections of test cases.

        The :class:`TestRef` object allows us to delay the creation of test runtime instance data and state until it is
        necessary to instantiate it and allows us to cleanup the runtime instance and state as soon as it is no longer
        being used.
    """

    def __init__(self, testfunc: FunctionType, monikers: List[str]=[], pivots: OrderedDict[str, Any]=collections.OrderedDict()):
        """
            Initializes the test reference object.

            :param testcontainer: The class of the test object that is being created.
            :param testmeth: The method on the test container
        """
        self._test_function = testfunc
        self._monikers = monikers
        self._pivots = pivots
        self._subscriptions = None
        self._finalized = False
        return

    @property
    def finalized(self) -> bool:
        return self._finalized

    @property
    def monikers(self) -> List[str]:
        return self._monikers

    @property
    def pivots(self) -> OrderedDict[str, Any]:
        return self._pivots

    @property
    def scope_name(self) -> str:
        return self.test_name

    @property
    def subscriptions(self):
        return self._subscriptions

    @subscriptions.setter
    def subscriptions(self, val):
        self._subscriptions = val
        return

    @property
    def test_base_name(self) -> str:
        tbname = self._test_function.__name__
        return tbname

    @property
    def test_function(self) -> FunctionType:
        """
            The test function 
        """
        return self._test_function

    @property
    def test_function_parameters(self):
        signature = inspect.signature(self._test_function)
        return signature.parameters

    @property
    def test_module_name(self) -> str:
        return self._test_function.__module__

    @property
    def test_name(self) -> str:
        """
            The fully qualified name of the test that is referenced.
        """
        tf = self._test_function
        test_name = "%s#%s" % (tf.__module__, tf.__name__)
        return test_name

    def finalize(self):
        self._finalized = True
        return

    def is_member_of_metaset(self, metafilters: Sequence[MetaFilter]):
        """
            Indicates if a test belongs to a set that is associated with a collection of metafilters.
        """
        include = True

        for mfilter in metafilters:
            if not mfilter.should_include(self._metadata):
                include = False

        return include

    def resolve_metadata(self, parent_metadata: Optional[Dict[str, str]]=None):

        reference_metadata = self._reference_metadata()

        if parent_metadata is not None:
            if reference_metadata is not None:
                self._metadata = {}
                self._metadata.update(parent_metadata)
                self._metadata.update(reference_metadata)
            else:
                self._metadata = parent_metadata
        else:
            self._metadata = reference_metadata

        return

    def _reference_metadata(self):
        """
            Looks up the metadata if any on the module associated with this group.
        """
        
        refmd = None
        if hasattr(self._test_function, "_metadata_"):
            refmd = self._test_function._metadata_

        return refmd

    def __str__(self):
        return self.test_name
